"""SC Controller - Action Editor - First Page.

Provides links for quick settings.
"""

import logging

from scc.actions import Action
from scc.gui.ae import AEComponent
from scc.tools import _, nameof

log = logging.getLogger("AE.1st")

__all__ = ["FirstPage"]

MARKUP_BUTTON = """
<big>%(what)s: Quick settings</big>

  • Map to <a href='page://buttons'>Button</a>

  • Use to <a href='quick://menu("Default.menu")'>display on-screen menu</a>
"""

MARKUP_TRIGGER = """
<big>%(what)s: Quick settings</big>

  • Map to <a href='quick://axis(Axes.ABS_Z)'>Left</a> or <a href='quick://axis(Axes.ABS_RZ)'>Right</a> Trigger

  • Map to <a href='quick://trigger(50, 255, button(Keys.BTN_LEFT))'>Left</a> or <a href='quick://trigger(50, 255, button(Keys.BTN_RIGHT))'>Right</a> mouse button

  • Map to <a href='grab://trigger_button'>Button</a>
"""


# TODO: Add haptics here
MARKUP_PAD = """
<big>%(what)s: Quick settings</big>

  • Use as <a href='quick://sens(1.2, 1.2, XY(axis(Axes.ABS_X), raxis(Axes.ABS_Y)))'>Left</a> or <a href='quick://sens(1.2, 1.2, XY(axis(Axes.ABS_RX), raxis(Axes.ABS_RY)))'>Right</a> gamepad stick, or <a href='quick://dpad(hatup(Axes.ABS_HAT0Y), hatdown(Axes.ABS_HAT0Y), hatleft(Axes.ABS_HAT0X), hatright(Axes.ABS_HAT0X))'>DPAD</a>

  • Use as keyboard: <a href='quick://dpad(button(Keys.KEY_W), button(Keys.KEY_S), button(Keys.KEY_A), button(Keys.KEY_D))'>WSAD</a> or <a href='quick://dpad(button(Keys.KEY_UP), button(Keys.KEY_DOWN), button(Keys.KEY_LEFT), button(Keys.KEY_RIGHT))'>Arrows</a>

  • Use as <a href='quick://smooth(8, 0.78, 2.0, feedback(RIGHT, 256, ball(mouse())))'>mouse</a> or <a href='quick://feedback(LEFT, 4096, 16.0, ball(XY(mouse(Rels.REL_HWHEEL), mouse(Rels.REL_WHEEL))))'>mouse wheel</a>

  • Create <a href='quick://feedback(LEFT, 4096, 16.0, menu("Default.menu",DEFAULT,A,B))'>touch menu</a>

  • Enable <a href='page://gesture'>gesture recognition</a>
"""

MARKUP_STICK = """
<big>Stick: Quick settings</big>

  • Use as <a href='quick://sens(1.2, 1.2, XY(axis(Axes.ABS_X), raxis(Axes.ABS_Y)))'>Left</a> or <a href='quick://sens(1.2, 1.2, XY(axis(Axes.ABS_RX), raxis(Axes.ABS_RY)))'>Right</a> gamepad stick, or <a href='quick://dpad(hatup(Axes.ABS_HAT0Y), hatdown(Axes.ABS_HAT0Y), hatleft(Axes.ABS_HAT0X), hatright(Axes.ABS_HAT0X))'>DPAD</a>

  • Use as keyboard: <a href='quick://dpad(button(Keys.KEY_W), button(Keys.KEY_S), button(Keys.KEY_A), button(Keys.KEY_D))'>WSAD</a> or <a href='quick://dpad(button(Keys.KEY_UP), button(Keys.KEY_DOWN), button(Keys.KEY_LEFT), button(Keys.KEY_RIGHT))'>Arrows</a>

  • Use as <a href='quick://smooth(8, 0.78, 2.0, feedback(RIGHT, 256, ball(mouse())))'>mouse</a> or <a href='quick://feedback(LEFT, 4096, 16.0, ball(XY(mouse(Rels.REL_HWHEEL), mouse(Rels.REL_WHEEL))))'>mouse wheel</a>
"""

MARKUP_GYRO = """
<big>Gyro: Quick settings</big>

  • Setup for aiming when right <a href='quick://mode(RPADTOUCH, mouse(ROLL), None)'>pad is touched</a>

  • Setup for aiming when right <a href='quick://mode(LT >= 0.7, mouse(ROLL), None)'>trigger is pushed</a>

  • <a href='quick://sens(3.5, 3.5, 3.5, mouse(ROLL))'>Use as mouse</a>
"""


class FirstPage(AEComponent):
	GLADE = "ae/first_page.glade"
	NAME = "first_page"
	CTXS = 0
	PRIORITY = 999

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)

	def load(self):
		if AEComponent.load(self):
			markup = ""
			if self.editor.get_mode() == Action.AC_PAD:
				markup = MARKUP_PAD
			elif self.editor.get_mode() == Action.AC_STICK:
				markup = MARKUP_STICK
			elif self.editor.get_mode() == Action.AC_GYRO:
				markup = MARKUP_GYRO
			elif self.editor.get_mode() == Action.AC_TRIGGER:
				markup = MARKUP_TRIGGER
			else:
				markup = MARKUP_BUTTON

			long_names = {
				"LPAD": _("Left Pad"),
				"RPAD": _("Right Pad"),
				"LGRIP": _("Left Grip"),
				"RGRIP": _("Right Grip"),
				"LB": _("Left Bumper"),
				"RB": _("Right Bumper"),
				"LEFT": _("Left Trigger"),
				"RIGHT": _("Right Trigger"),
				"STICK": _("Stick"),
			}

			markup = markup % {
				"what": long_names.get(nameof(self.editor.get_id()), nameof(self.editor.get_id()).title()),
			}
			self.builder.get_object("lblMarkup").set_markup(markup.strip(" \r\n\t"))
			return True

	def on_lblMarkup_activate_link(self, trash, link):
		self.editor.on_link(link)



"""SC-Controller - About dialog."""

from scc.gui.editor import Editor


class AboutDialog(Editor):
	"""Standard looking about dialog."""

	GLADE = "about.glade"

	def __init__(self, app) -> None:
		self.app = app
		self.setup_widgets()

	def setup_widgets(self) -> None:
		"""Setup widgets and get and set app version in the About.

		TODO - We can get a mismatch like this and version won't be set then, why?:
			location:     /usr/lib/python3.12/site-packages
			scc.__file__: /home/user/sc-controller/scc/__init__.py
		"""
		Editor.setup_widgets(self)

		app_ver = "(unknown version)"
		import importlib.metadata
		import importlib.resources

		import scc

		sccontroller_version = importlib.metadata.version("sccontroller")
		sccontroller_module = importlib.resources.files("scc")
		if sccontroller_module is not None:
			if scc.__file__.startswith(str(sccontroller_module)):
				app_ver = "v" + sccontroller_version

			else:
				print(
					"Could not get version, locations possibly could not match.",
					f"scc.__file__:{scc.__file__}",
					f"sccontroller_module: {sccontroller_module}",
					f"sccontroller_version: {sccontroller_version}",
				)
		# Display version in UI
		self.builder.get_object("lblVersion").set_label(app_ver)

	def show(self, modal_for) -> None:
		if modal_for:
			self.window.set_transient_for(modal_for)
			self.window.set_modal(True)
		self.window.show()

	def on_dialog_response(self, *a) -> None:
		self.close()



#!/usr/bin/env python3
"""SC-Controller - Controller Image

Big, SVGWidget based widget with interchangeable controller and button images.
"""

import copy
import json
import logging
import os

from scc.constants import SCButtons
from scc.gui.svg_widget import SVGEditor, SVGWidget
from scc.tools import nameof

log = logging.getLogger("ContImage")


class ControllerImage(SVGWidget):
	DEFAULT = "sc"
	BUTTONS_WITH_IMAGES = (
		SCButtons.A,
		SCButtons.B,
		SCButtons.X,
		SCButtons.Y,
		SCButtons.BACK,
		SCButtons.C,
		SCButtons.START,
		SCButtons.DOTS,
	)

	DEFAULT_AXES = (
		# Shared between DS4 and Steam Controller
		"stick_x",
		"stick_y",
		"lpad_x",
		"lpad_x",
		"rpad_y",
		"rpad_y",
		"ltrig",
		"rtrig",
	)

	DEFAULT_BUTTONS = [nameof(x) for x in BUTTONS_WITH_IMAGES] + [
		# Used only by Steam Controller
		nameof(SCButtons.LB),
		nameof(SCButtons.RB),
		nameof(SCButtons.LT),
		nameof(SCButtons.RT),
		nameof(SCButtons.STICKPRESS),
		nameof(SCButtons.RPAD),
		nameof(SCButtons.LPAD),
		nameof(SCButtons.LGRIP),
		nameof(SCButtons.RGRIP),
	]

	def __init__(self, app, config=None):
		self.app = app
		self.backup = None
		self.current = self._ensure_config({}, None)
		filename = self._make_controller_image_path(ControllerImage.DEFAULT)
		SVGWidget.__init__(self, filename)
		if config:
			self._controller_image.use_config(config)

	def _make_controller_image_path(self, img):
		return os.path.join(self.app.imagepath, f"controller-images/{img}.svg")

	def get_config(self):
		"""Returns last used config
		"""
		return self.current

	def _ensure_config(self, data, controller):
		"""Ensure that required keys are present in config data"""
		data["gui"] = data.get("gui", {})
		data["gui"]["background"] = data["gui"].get("background", "sc")
		data["gui"]["buttons"] = data["gui"].get("buttons") or self._get_default_images()
		data["gui"]["no_buttons_in_gui"] = data["gui"].get("no_buttons_in_gui") or False
		data["buttons"] = data.get("buttons") or ControllerImage.DEFAULT_BUTTONS
		data["axes"] = data.get("axes") or ControllerImage.DEFAULT_AXES
		data["gyros"] = data.get("gyros", data["gui"]["background"] == "sc")
		return data

	@staticmethod
	def get_names(dict_or_tuple):
		"""There are three different ways how button and axis names are stored
		in config. This wrapper provides unified way to get list of them.
		"""
		if type(dict_or_tuple) in (list, tuple):
			return dict_or_tuple
		return [(x["axis"] if type(x) is dict else x) for x in dict_or_tuple.values()]

	def use_config(self, config, backup=None, controller=None):
		"""Loads controller settings from provided config, adding default values
		when needed. Returns same config.
		"""
		self.backup = backup
		self.current = self._ensure_config(config or {}, controller)
		self.set_image(os.path.join(self.app.imagepath, f"controller-images/{self.current['gui']['background']}.svg"))
		if not self.current["gui"]["no_buttons_in_gui"]:
			self._fill_button_images(self.current["gui"]["buttons"])
		self.hilight({})
		return self.current

	def override_background(self, filename):
		"""Overrides background image setting. This changes config in place,
		so next time get_config is called, changed background is part of it.
		"""
		if self.backup is None:
			self.backup = copy.deepcopy(self.current)
		data = json.loads(open(os.path.join(self.app.imagepath, f"{filename}.json")).read())
		self.current["gui"]["background"] = data["gui"]["background"]
		self.use_config(self.current, self.backup)

	def override_buttons(self, filename):
		"""Overrides button settings. This changes config in place,
		so next time get_config is called, changed background is part of it.
		"""
		if self.backup is None:
			self.backup = copy.deepcopy(self.current)
		data = json.loads(open(os.path.join(self.app.imagepath, "{filename}.json")).read())
		self.current["gui"]["buttons"] = data["gui"]["buttons"]
		self.current["buttons"] = data["buttons"]
		self.use_config(self.current, self.backup)

	def undo_override(self):
		"""Undoes override_* changes"""
		if self.backup is not None:
			self.use_config(self.backup, None)

	def get_button_groups(self):
		groups = json.loads(open(os.path.join(self.app.imagepath, "button-images", "groups.json")).read())
		return {x["key"]: x["buttons"] for x in groups if x["type"] == "buttons"}

	def _get_default_images(self):
		return self.get_button_groups()[ControllerImage.DEFAULT]

	def _fill_button_images(self, buttons):
		e = self.edit()
		# SVGEditor.update_parents(e)
		target = SVGEditor.get_element(e, "controller")
		target_x, target_y = SVGEditor.get_translation(target)
		for i in range(len(ControllerImage.BUTTONS_WITH_IMAGES)):
			b = nameof(ControllerImage.BUTTONS_WITH_IMAGES[i])
			if b == "DOTS":
				# How did I managed to create this kind of special case? -_-
				i = 16
			path = None
			try:
				elm = SVGEditor.get_element(e, f"AREA_{b}")
				if elm is None:
					log.warning(f"Area for button {b} not found")
					continue
				x, y = SVGEditor.get_translation(elm)
				scale = 1.0
				if "scc-button-scale" in elm.attrib:
					w, h = SVGEditor.get_size(elm)
					scale = float(elm.attrib["scc-button-scale"])
					tw, th = w * scale, h * scale
					if scale < 1.0:
						x += (w - tw) * 0.5
						y += (h - th) * 0.5
					else:
						x -= (tw - w) * 0.25
						y -= (th - h) * 0.25
				path = os.path.join(self.app.imagepath, "button-images", f"{buttons[i]}.svg")
				img = SVGEditor.get_element(SVGEditor.load_from_file(path), "button")
				img.attrib["transform"] = f"translate({x - target_x}, {y - target_y}) scale({scale})"
				img.attrib["id"] = b
				SVGEditor.add_element(target, img)

			except Exception as err:
				log.warning(f"Failed to add image for button {b} (from {path})")
				log.exception(err)
		e.commit()



"""SC-Controller - Daemon - CemuHookUDP motion provider.

Accepts all connections from clients and sends data captured
by 'cemuhook' actions to them.
"""

import logging
import os
import socket
from ctypes import c_bool, c_char_p, c_float, c_int, c_size_t, create_string_buffer
from datetime import datetime, timedelta
from enum import IntEnum
from threading import Thread
from time import sleep
from typing import TYPE_CHECKING

from scc.tools import find_library

if TYPE_CHECKING:
	from ctypes import CDLL

log = logging.getLogger("CemuHook")

BUFFER_SIZE = 1024
IP = "127.0.0.1"
PORT = 26760


class MessageType(IntEnum):
	DSUC_VERSIONREQ = 0x100000
	DSUS_VERSIONRSP = 0x100000
	DSUC_LISTPORTS = 0x100001
	DSUS_PORTINFO = 0x100001
	DSUC_PADDATAREQ = 0x100002
	DSUS_PADDATARSP = 0x100002


class CemuhookServer:
	C_DATA_T = c_float * 6
	timeout = timedelta(seconds=1)

	def __init__(self, daemon):
		self._lib: CDLL = find_library("libcemuhook")
		self._lib.cemuhook_data_received.argtypes = [c_int, c_char_p, c_int, c_char_p, c_size_t]
		self._lib.cemuhook_data_received.restype = None
		self._lib.cemuhook_feed.argtypes = [c_int, c_int, CemuhookServer.C_DATA_T]
		self._lib.cemuhook_feed.restype = None
		self._lib.cemuhook_socket_enable.argtypes = []
		self._lib.cemuhook_socket_enable.restype = c_bool
		self.last_signal: datetime = datetime.now()

		if not self._lib.cemuhook_socket_enable():
			raise OSError("cemuhook_socket_enable failed")

		self.socket: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

		poller = daemon.get_poller()
		daemon.poller.register(self.socket.fileno(), poller.POLLIN, self.on_data_received)

		server_port = int(os.getenv("SCC_SERVER_PORT") or PORT)
		server_ip = os.getenv("SCC_SERVER_IP") or IP
		self.socket.bind((server_ip, server_port))
		log.info("Created CemuHookUDP Motion Provider")

		Thread(target=self._keepalive).start()

	def _keepalive(self):
		while True:
			if datetime.now() - self.last_signal >= self.timeout:
				# feed all zeroes to indicate the gyro has not changed
				self.feed((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
			sleep(1)

	def on_data_received(self, fd: int, event_type):
		if fd != self.socket.fileno():
			return
		message, (ip, port) = self.socket.recvfrom(BUFFER_SIZE)
		buffer = create_string_buffer(BUFFER_SIZE)
		self._lib.cemuhook_data_received(fd, ip.encode("utf-8"), port, message, len(message), buffer)

	def feed(self, data):
		self.last_signal = datetime.now()
		c_data = CemuhookServer.C_DATA_T()
		# log.debug(data)
		c_data[0:6] = data[0:6]
		# log.debug(list(c_data))
		self._lib.cemuhook_feed(self.socket.fileno(), 0, c_data)



"""SC Controller - Controller Registration data

Dummy container classes
"""

import logging

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN
from scc.gui.creg.constants import AXIS_TO_BUTTON

log = logging.getLogger("CReg.data")


class AxisData:
	"""(Almost) dumb container.

	Stores position, center and limits for single axis.
	"""

	def __init__(self, name, xy, min=STICK_PAD_MAX, max=STICK_PAD_MIN):
		self.name = name
		self.area = name.split("_")[0].upper()
		if self.area.endswith("TRIG"):
			self.area = self.area[0:-3]
		self.xy = xy
		self.pos = 0
		self.center = 0
		self.min = min
		self.max = max
		self.invert = False
		self.cursor = None

	def reset(self):
		"""Reset min and max value so axis can (has to be) recalibrated again"""
		self.min = STICK_PAD_MAX
		self.max = STICK_PAD_MIN

	def __repr__(self):
		return "<Axis data '%s'>" % (self.name,)

	def set_position(self, value):
		"""Return (changed, x), value determining if axis limits were changed and current position position.

		translated to range of (STICK_PAD_MIN, STICK_PAD_MAX)
		"""
		changed = False
		if value < self.min:
			self.min = value
			changed = True
		if value > self.max:
			self.max = value
			changed = True
		self.pos = value
		try:
			r = (STICK_PAD_MAX - STICK_PAD_MIN) / (self.max - self.min)
			v = (self.pos - self.min) * r
			if self.invert:
				return changed, STICK_PAD_MAX - v
			return changed, v + STICK_PAD_MIN
		except ZeroDivisionError:
			return changed, 0


class DPadEmuData:
	"""Dumb container that stores dpad emulation data.

	DPAd emulation is used, for example, on PS3 controller, where dpad does not
	inputs as 2 axes, but as 4 buttons.

	This class stores mapping of one button to one half of axis.
	"""

	def __init__(self, axis_data, positive):
		self.axis_data = axis_data
		self.positive = positive
		self.button = AXIS_TO_BUTTON[axis_data.name]



#!/usr/bin/env python3
"""SC-Controller - OSD Launcher

Display launcher with phone-like keyboard that user can use to select
application (list is generated using xdg) and start it.

Reuses styles from OSD Menu and OSD Dialog
"""

import logging
import os

from gi.repository import GdkX11, Gio, Gtk, Pango

from scc.config import Config
from scc.constants import DEFAULT, LEFT, RIGHT, STICK, STICK_PAD_MAX
from scc.gui.daemon_manager import DaemonManager
from scc.lib import xwrappers as X
from scc.osd import OSDWindow, StickController
from scc.paths import get_share_path
from scc.tools import _, circle_to_square, clamp, point_in_gtkrect

log = logging.getLogger("osd.menu")


class Launcher(OSDWindow):
	EPILOG = """Exit codes:
   0  - clean exit, user selected option
  -1  - clean exit, user canceled dialog
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while dialog is displayed.
   3  - erorr, failed to lock input stick, pad or button(s)
	"""

	BUTTONS = ["1", "2 ABC", "3 DEF", "5 GHI", "5 JKL", "6 MNO", "7 PQRS", "8 TUV", "9 WXYZ", "", "0"]

	VALID_CHARS = "12ABC3DEF5GHI5JKL6MNO7PQRS8TUV9WXYZ0"
	CHAR_TO_NUMBER = {}  # Generated on runtime

	MAX_ROWS = 5

	_app_db = None  # Static list of all know applications

	def __init__(self, cls="osd-menu"):
		self._buttons = None
		self._string = ""

		OSDWindow.__init__(self, cls)
		self.daemon = None
		self.config = None
		self.feedback = None
		self.controller = None
		self.xdisplay = X.Display(hash(GdkX11.x11_get_default_xdisplay()))  # Magic

		self.create_parent()
		self.create_app_list()
		self.create_buttons()

		cursor = os.path.join(get_share_path(), "images", "menu-cursor.svg")
		self.cursors = [Gtk.Image.new_from_file(cursor), Gtk.Image.new_from_file(cursor)]
		for c in self.cursors:
			c.set_name("osd-menu-cursor")
			c.selected = None
			self.f.add(c)
		self.f.show_all()

		self._scon = StickController()
		self._scon.connect("direction", self.on_stick_direction)
		self._selected = None
		self._menuid = None
		self._eh_ids = []
		self._confirm_with = "A"
		self._cancel_with = "B"

		if Launcher._app_db is None:
			Launcher._app_db = []
			for x in Launcher.BUTTONS:
				for c in x:
					Launcher.CHAR_TO_NUMBER[c] = x[0]

			for x in Gio.AppInfo.get_all():
				try:
					Launcher._app_db.append((Launcher.name_to_keys(x), x))
				except UnicodeDecodeError:
					# Just fuck them...
					pass

	@staticmethod
	def name_to_keys(appinfo):
		name = "".join(
			[Launcher.CHAR_TO_NUMBER[x] for x in appinfo.get_display_name().upper() if x in Launcher.VALID_CHARS],
		)
		return name

	@staticmethod
	def string_to_keys_and_spaces(string):
		name = "".join([Launcher.CHAR_TO_NUMBER[x] if x in Launcher.VALID_CHARS else " " for x in string.upper()])
		return name

	def create_parent(self):
		self.parent = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		self.parent.set_name("osd-dialog")
		self.f = Gtk.Fixed()
		self.f.add(self.parent)
		self.add(self.f)

	def create_app_list(self):
		lst = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		lst.set_name("osd-application-list")
		self.items = [self.generate_widget("") for x in range(self.MAX_ROWS)]
		for a in self.items:
			lst.pack_start(a, False, True, 0)
		self.parent.pack_start(lst, True, True, 0)
		self._set_launchers([])
		lst.show_all()

	def create_buttons(self):
		self.grid = Gtk.Grid()
		self.parent.pack_start(self.grid, True, True, 0)
		self._buttons = []

		x, y = 0, 0
		for label in self.BUTTONS:
			if label:
				w = self.generate_widget(label)
				w.set_name("osd-key-buton")
				self.grid.attach(w, x, y, 1, 1)
				self._buttons.append(w)
			x += 1
			if x > 2:
				x = 0
				y += 1

		w = self.generate_widget(_("Run"))
		self.grid.attach(w, x, y, 1, 1)

		self.grid.set_name("osd-dialog-buttons")

	def pack_items(self, parent, items):
		for item in items:
			if hasattr(item.widget, "set_alignment"):
				item.widget.set_alignment(0.5, 0.5)
			self._buttons.pack_end(item.widget, True, True, 0)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process.
		use_config() should be be called before parse_argumets() if this is used.
		"""
		self.daemon = d
		self._connect_handlers()
		self.on_daemon_connected(self.daemon)

	def use_config(self, c):
		"""Allows reusing already existin Config instance in same process.
		Has to be called before parse_argumets()
		"""
		self.config = c

	def get_menuid(self):
		"""Returns ID of used menu.
		"""
		return

	def get_selected_item_id(self):
		"""Returns ID of selected item or None if nothing is selected.
		"""
		return

	def _launch(self):
		self._selected.launcher.launch()

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"--confirm-with", type=str, metavar="button", default=DEFAULT, help="button used to confirm choice",
		)
		self.argparser.add_argument(
			"--cancel-with", type=str, metavar="button", default=DEFAULT, help="button used to cancel dialog",
		)
		self.argparser.add_argument(
			"--feedback-amplitude",
			type=int,
			help="enables and sets power of feedback effect generated when active menu option is changed",
		)

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		if not self.config:
			self.config = Config()

		if self.args.feedback_amplitude:
			side = "LEFT"
			self.feedback = side, int(self.args.feedback_amplitude)

		# Create buttons that are displayed on screen
		return True

	def _set_launchers(self, launchers):
		launchers = launchers[0 : self.MAX_ROWS]
		for x in self.items:
			x.set_label("")
			x.set_name("osd-hidden-item")
			x.launcher = None
		for i in range(len(launchers)):
			self.items[i].set_name("osd-launcher-item")
			self.items[i].launcher = launchers[i]
			label = self.items[i].get_children()[0]
			label.set_markup(self._format_label_markup(launchers[i]))
			label.set_max_width_chars(1)
			label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
			label.set_xalign(0)

	def _format_label_markup(self, label):
		if hasattr(label, "get_display_name"):
			label = label.get_display_name()
		else:
			label = str(label)

		def _check(substr):
			i, ch = 0, self._string
			while len(substr) > 0 and substr[0] == ch[0]:
				ch = ch[1:]
				substr = substr[1:]
				i += 1
				if len(ch) == 0:
					return i
			while len(substr) > 0 and substr[0] == " ":
				substr = substr[1:]
				i += 1
			return -1

		keys = Launcher.string_to_keys_and_spaces(label)
		index1, index2 = -1, -1
		for i in range(len(keys)):
			if keys[i] == self._string[0]:
				index2 = _check(keys[i:])
				if index2 > 0:
					index1 = i
					index2 = i + index2
					break

		label = "%s<span color='#%s'>%s</span>%s" % (
			label[0:index1],
			self.config["osd_colors"]["menuitem_hilight_text"],
			label[index1:index2],
			label[index2:],
		)
		return label

	def _update_items(self):
		if len(self._string) > 0:
			gen = (item for (keys, item) in self._app_db if self._string in keys)
			launchers = []
			for i in gen:
				launchers.append(i)
				if len(launchers) > self.MAX_ROWS:
					break
			self._set_launchers(launchers)
			self.select(0)
		else:
			self._set_launchers([])

	def generate_widget(self, label):
		"""Generates gtk widget for specified menutitem"""
		if hasattr(label, "label"):
			label = label.label
		widget = Gtk.Button.new_with_label(label)
		widget.set_relief(Gtk.ReliefStyle.NONE)
		if hasattr(widget.get_children()[0], "set_xalign"):
			widget.get_children()[0].set_xalign(0)
		else:
			widget.get_children()[0].set_halign(Gtk.Align.START)
		widget.set_name("osd-menu-item")

		return widget

	def select(self, index):
		if self._selected:
			self._selected.set_name(self._selected.get_name().replace("-selected", ""))
			self._selected = None
		if self.items[index].launcher is not None:
			self._selected = self.items[index]
			self._selected.set_name(self._selected.get_name() + "-selected")
			return True
		return False

	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)

	def show(self, *a):
		if not self.select(0):
			self.next_item(1)
		OSDWindow.show(self, *a)
		for c in self.cursors:
			c.set_visible(False)

	def on_daemon_connected(self, *a):
		def success(*a):
			log.error("Sucessfully locked input")

		if not self.config:
			self.config = Config()
		self.controller = self.choose_controller(self.daemon)
		if self.controller is None or not self.controller.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		ccfg = self.config.get_controller_config(self.controller.get_id())
		self._confirm_with = ccfg["menu_confirm"] if self.args.confirm_with == DEFAULT else self.args.confirm_with
		self._cancel_with = ccfg["menu_cancel"] if self.args.cancel_with == DEFAULT else self.args.cancel_with

		self._eh_ids += [
			(self.controller, self.controller.connect("event", self.on_event)),
			(self.controller, self.controller.connect("lost", self.on_controller_lost)),
		]
		locks = [LEFT, RIGHT, STICK, "LPAD", "RPAD", "LB", self._confirm_with, self._cancel_with]
		self.controller.lock(success, self.on_failed_to_lock, *locks)

	def quit(self, code=-2):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		OSDWindow.quit(self, code)

	def next_item(self, direction):
		"""Selects next menu item, based on self._direction"""
		start, i = -1, 0
		try:
			start = self.items.index(self._selected)
			i = start + direction
		except:
			pass
		while True:
			if i == start:
				# Cannot find valid menu item
				self.select(start)
				break
			if i >= len(self.items):
				i = 0
				continue
			if i < 0:
				i = len(self.items) - 1
				continue
			if self.select(i):
				break
			i += direction
			start = max(start, 0)

	def on_stick_direction(self, trash, x, y):
		if y != 0:
			self.next_item(y)

	def _move_cursor(self, cursor, x, y):
		if (x, y) == (0, 0):
			cursor.set_visible(False)
			return
		cursor.set_visible(True)
		pad_w = cursor.get_allocation().width * 0.5
		pad_h = cursor.get_allocation().height * 0.5
		max_w = self.grid.get_allocation().width - 2 * pad_w
		max_h = self.grid.get_allocation().height - 2 * pad_h

		x, y = circle_to_square(x / (STICK_PAD_MAX * 2.0), y / (STICK_PAD_MAX * 2.0))
		x = clamp(pad_w, (pad_w + max_w) * 0.5 + x * max_w, max_w - pad_w)
		y = clamp(pad_h, (pad_h + max_h) * 0.5 + y * max_h * -1, max_h - pad_h)
		x += self.grid.get_allocation().x
		y += self.grid.get_allocation().y
		self.f.move(cursor, int(x), int(y))

		for i in self._buttons:
			if point_in_gtkrect(i.get_allocation(), x, y):
				if cursor.selected:
					cursor.selected.set_name("osd-key-buton")
				cursor.selected = i
				cursor.selected.set_name("osd-key-buton-hilight")
				break

	def _get_under_cursor(self, cursor):
		x, y = self.f.child_get(cursor, "x", "y")
		for i in self._buttons:
			if point_in_gtkrect(i.get_allocation(), x, y):
				return i
		return None

	def on_event(self, daemon, what, data):
		if what == LEFT:
			self._move_cursor(self.cursors[0], *data)
		elif what == "LPAD" and data[0] == 1:
			b = self._get_under_cursor(self.cursors[0])
			if b:
				self._string += b.get_label()[0]
			self._update_items()
		elif what == RIGHT:
			self._move_cursor(self.cursors[1], *data)
		elif what == "RPAD" and data[0] == 1:
			b = self._get_under_cursor(self.cursors[1])
			if b:
				self._string += b.get_label()[0]
			self._update_items()
		elif what == "LB":
			if len(self._string) > 0:
				self._string = self._string[:-1]
				self._update_items()
		elif what == STICK:
			self._scon.set_stick(*data)
		elif what == self._cancel_with:
			if data[0] == 0:  # Button released
				self.quit(-1)
		elif what == self._confirm_with:
			if data[0] == 0:  # Button released
				if self._selected:
					self._launch()
					self.quit(0)
				else:
					self.quit(-1)



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2015 Stany MARCEL <stanypub@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import ast
import operator as op
import os
import shlex
from collections import OrderedDict

OPERATORS = {
	ast.Add: op.add,
	ast.Sub: op.sub,
	ast.Mult: op.mul,
	ast.Div: op.floordiv,
	ast.Mod: op.mod,
	ast.LShift: op.lshift,
	ast.RShift: op.rshift,
	ast.BitOr: op.or_,
	ast.BitXor: op.xor,
	ast.BitAnd: op.and_,
	ast.Invert: op.invert,
	ast.Not: op.not_,
	ast.UAdd: op.pos,
	ast.USub: op.neg,
	ast.And: op.and_,
	ast.Or: op.or_,
	ast.Eq: op.eq,
	ast.NotEq: op.ne,
	ast.Lt: op.lt,
	ast.LtE: op.le,
	ast.Gt: op.gt,
	ast.GtE: op.ge,
}


def eval_expr(expr):
	"""Eval and expression inside a #define using a suppart of python grammar"""

	def _eval(node):
		if isinstance(node, ast.Constant):
			return node.value
		if isinstance(node, ast.BinOp):
			return OPERATORS[type(node.op)](_eval(node.left), _eval(node.right))
		if isinstance(node, ast.UnaryOp):
			return OPERATORS[type(node.op)](_eval(node.operand))
		if isinstance(node, ast.BoolOp):
			values = [_eval(x) for x in node.values]
			return OPERATORS[type(node.op)](**values)
		raise TypeError(node)

	return _eval(ast.parse(expr, mode="eval").body)


def defines(base, include):
	"""Extract #define from base/include following #includes"""
	parsed = set()
	fname = os.path.normpath(os.path.abspath(os.path.join(base, include)))
	parsed.add(fname)

	with open(fname) as file:
		lexer = shlex.shlex(file, posix=True)

		lexer.whitespace = " \t\r"
		lexer.commenters = ""
		lexer.quotes = '"'

		out = OrderedDict()

		def parse_c_comments(lexer, tok, ntok):
			if tok != "/" or ntok != "*":
				return False
			quotes = lexer.quotes
			lexer.quotes = ""
			while True:
				tok = lexer.get_token()
				ntok = lexer.get_token()
				if tok == "*" and ntok == "/":
					lexer.quotes = quotes
					break
				lexer.push_token(ntok)
			return True

		def parse_cpp_comments(lexer, tok, ntok):
			if tok != "/" or ntok != "/":
				return False
			quotes = lexer.quotes
			lexer.quotes = ""
			while True:
				tok = lexer.get_token()
				if tok == "\n":
					lexer.quotes = quotes
					lexer.push_token(tok)
					break
			return True

		while True:
			tok = lexer.get_token()
			if not tok or tok == "":
				break
			ntok = lexer.get_token()

			if parse_c_comments(lexer, tok, ntok):
				continue
			if parse_cpp_comments(lexer, tok, ntok):
				continue

			if tok != "\n" or ntok != "#":
				lexer.push_token(ntok)
				continue

			tok = lexer.get_token()
			if tok == "define":
				name = lexer.get_token()
				expr = ""
				while True:
					tok = lexer.get_token()
					ntok = lexer.get_token()

					if parse_c_comments(lexer, tok, ntok):
						continue
					if parse_cpp_comments(lexer, tok, ntok):
						continue
					lexer.push_token(ntok)

					if not tok or tok == "":
						break
					if tok == "\n":
						lexer.push_token(tok)
						break

					if tok in out:
						tok = str(out[tok])
					expr = expr + tok

				try:
					val = eval_expr(expr)
					out[name] = val
				except (SyntaxError, TypeError):
					pass
			elif tok == "include":
				tok = lexer.get_token()
				if tok == "<":
					name = ""
					while True:
						tok = lexer.get_token()
						if tok == ">":
							break
						name = name + tok
				else:
					name = tok
				fname = os.path.normpath(os.path.abspath(os.path.join(base, name)))
				if os.path.isfile(fname) and fname not in parsed:
					parsed.add(fname)
					lexer.push_source(open(fname))
			else:
				lexer.push_token(tok)

		return out


if __name__ == "__main__":
	import sys

	definesDict = defines(sys.argv[1], sys.argv[2])
	for k, v in definesDict.items():
		print(f"{k}:\t{v}")



"""SC-Controller - Import / Export Dialog
"""

import json
import logging
import tarfile
import traceback

from scc.gui.editor import ComboSetter, Editor
from scc.tools import _, find_profile, profile_is_default, profile_is_override

from .export import Export
from .import_sccprofile import ImportSccprofile
from .import_vdf import ImportVdf

log = logging.getLogger("IE.Dialog")


class Dialog(Editor, ComboSetter, Export, ImportVdf, ImportSccprofile):
	GLADE = "import_export.glade"

	def __init__(self, app):
		self.app = app
		self._back = []
		self._recursing = False
		self._next_callback = None
		self.setup_widgets()
		Export.__init__(self)
		ImportVdf.__init__(self)
		ImportSccprofile.__init__(self)

	@staticmethod
	def determine_type(filename):
		"""Detects and returns type of passed file, if it can be imported.
		Returns one of 'sccprofile', 'sccprofile.tar.gz', 'vdf', 'vdffz'
		or None if type is not supported.
		"""
		try:
			with open(filename, "rb") as file:
				f = file.read(1024)
		except Exception:
			# File not readable
			log.error(traceback.format_exc())
			return None
		try:
			if f.decode("utf-8").strip(" \t\r\n").startswith("{"):
				# Looks like json
				data = json.loads(open(filename).read())
				if "buttons" in data and "gyro" in data:
					return "sccprofile"
				if "GameName" in data and "FileName" in data:
					return "vdffz"
		except:
			# Definitelly not json
			pass

		if f[0:2] == b"\x1f\x8b":
			# gzip, hopefully tar.gz
			try:
				with tarfile.open(filename, "r:gz") as tar:
					names = [x.name for x in tar]
				any_profile = any([x.endswith(".sccprofile") for x in names])
				if any_profile and "profile-name" in names:
					return "sccprofile.tar.gz"
			except:
				# Not a tarball
				pass

		# Rest is decided by extension
		if filename.endswith(".sccprofile.tar.gz"):
			return "sccprofile.tar.gz"
		if filename.endswith(".vdf"):
			return "vdf"
		# Fallbacks if above fails
		if filename.endswith(".sccprofile"):
			return "sccprofile"
		if filename.endswith(".vdffz"):
			return "vdffz"
		return None

	@staticmethod
	def check_name(name):
		if len(name.strip()) <= 0:
			return False
		if "/" in name:
			return False
		if find_profile(name):
			# Profile already exists
			if profile_is_default(name) and not profile_is_override(name):
				# Existing profile is default and has no override yet
				return True
			return False
		return True

	def import_file(self, filename, filetype=None):
		"""Attempts to import passed file.

		Switches to apropriate page automatically, or, if file cannot be
		imported, does nothing.
		"""
		filetype = filetype or Dialog.determine_type(filename)
		if filetype == "sccprofile":
			self.import_scc(filename=filename)
		elif filetype == "sccprofile.tar.gz":
			self.import_scc_tar(filename=filename)
		elif filetype in ("vdf", "vdffz"):
			self.import_vdf(filename=filename)

	def next_page(self, page):
		stDialog = self.builder.get_object("stDialog")
		btBack = self.builder.get_object("btBack")
		self._back.append(stDialog.get_visible_child())
		stDialog.set_visible_child(page)
		btBack.set_visible(True)
		self._page_selected(page)

	def _page_selected(self, page):
		stDialog = self.builder.get_object("stDialog")
		hbDialog = self.builder.get_object("hbDialog")
		hbDialog.set_title(stDialog.child_get_property(page, "title"))
		hname = "on_%s_activated" % (page.get_name(),)
		if hasattr(self, hname):
			getattr(self, hname)()

	def enable_next(self, enabled=True, callback=None):
		"""Makes 'Next' button visible and assigns callback that will be
		called when button is clicked. 'Next' button is automatically hidden
		before callback is called.

		Returns 'Next' button widget.
		"""
		assert not enabled or callback
		btNext = self.builder.get_object("btNext")
		btNext.set_visible(enabled)
		btNext.set_use_stock(False)
		btNext.set_sensitive(True)
		btNext.set_label(_("Next"))
		self._next_callback = callback
		return btNext

	def on_btNext_clicked(self, *a):
		cb = self._next_callback
		self.enable_next(enabled=False)
		cb()

	def on_btBack_clicked(self, *a):
		btBack = self.builder.get_object("btBack")
		stDialog = self.builder.get_object("stDialog")
		btSaveAs = self.builder.get_object("btSaveAs")
		btNext = self.builder.get_object("btNext")
		page, self._back = self._back[-1], self._back[:-1]
		stDialog.set_visible_child(page)
		btNext.set_visible(False)
		btSaveAs.set_visible(False)
		btBack.set_visible(len(self._back) > 0)
		self._page_selected(page)

	def on_btExport_clicked(self, *a):
		grSelectProfile = self.builder.get_object("grSelectProfile")
		self.next_page(grSelectProfile)

	def on_btImportVdf_clicked(self, *a):
		grVdfImport = self.builder.get_object("grVdfImport")
		self.next_page(grVdfImport)



import logging
import os
import tarfile

from gi.repository import GLib, GObject, Gtk

from scc.gui.parser import GuiActionParser
from scc.menu_data import MenuData
from scc.profile import Encoder, Profile
from scc.special_actions import ChangeProfileAction, MenuAction, ShellCommandAction
from scc.tools import _, get_menus_path, get_profiles_path

from .export import Export

log = logging.getLogger("IE.ImportSSCC")


class ImportSccprofile:
	def on_btImportSccprofile_clicked(self, *a):
		# Create filters
		f1 = Gtk.FileFilter()
		f1.set_name("SC Controller Profile or Archive")
		f1.add_pattern("*.sccprofile")
		f1.add_pattern("*.sccprofile.tar.gz")

		# Create dialog
		d = Gtk.FileChooserNative.new(_("Import Profile..."), self.window, Gtk.FileChooserAction.OPEN)
		d.add_filter(f1)
		if d.run() == Gtk.ResponseType.ACCEPT:
			if d.get_filename().endswith(".tar.gz"):
				self.import_scc_tar(d.get_filename())
			else:
				self.import_scc(d.get_filename())

	def error(self, text):
		"""Displays error page (reused from VDF import).
		"""
		tbError = self.builder.get_object("tbError")
		grImportFailed = self.builder.get_object("grImportFailed")

		tbError.set_text(text)
		self.next_page(grImportFailed)

	def import_scc(self, filename):
		"""Imports simple, single-file scc-profile.
		Just loads it, checks for shell() actions and asks user to enter name.
		"""
		files = self.builder.get_object("lstImportPackage")
		# Load profile
		profile = Profile(GuiActionParser())
		try:
			profile.load(filename)
		except Exception as e:
			# Profile cannot be parsed. Display error message and let user to quit
			# Error message reuses page from VDF import, because they are
			# basically the same
			log.error(e)
			self.error(str(e))
			return

		name = ".".join(os.path.split(filename)[-1].split(".")[0:-1])
		files.clear()
		o = GObject.GObject()
		o.obj = profile
		files.append((2, name, name, _("(profile)"), o))

		self.check_shell_commands()

	def import_scc_tar(self, filename):
		"""Imports packaged profiles.
		Checks for shell() actions everywhere and ask user to
		enter main name, check generated ones and optionaly change
		them as he wish.
		"""
		files = self.builder.get_object("lstImportPackage")
		try:
			# Open tar
			tar = tarfile.open(filename, "r:gz")
			files.clear()
			# Grab 1st profile
			name = tar.extractfile(Export.PN_NAME).read()
			main_profile = "%s.sccprofile" % name
			parser = GuiActionParser()
			o = GObject.GObject()
			o.obj = Profile(parser).load_fileobj(tar.extractfile(main_profile))
			files.append((2, name, name, _("(profile)"), o))
			for x in tar:
				name = ".".join(x.name.split(".")[0:-1])
				if x.name.endswith(".sccprofile") and x.name != main_profile:
					o = GObject.GObject()
					o.obj = Profile(parser).load_fileobj(tar.extractfile(x))
					files.append((True, name, name, _("(profile)"), o))
				elif x.name.endswith(".menu"):
					o = GObject.GObject()
					o.obj = MenuData.from_fileobj(tar.extractfile(x), parser)
					files.append((True, name, name, _("(menu)"), o))
		except Exception as e:
			# Either entire tar or some profile cannot be parsed.
			# Display error message and let user to quit
			# Error message reuses same page as above.
			log.error(e)
			self.error(str(e))
			return
		self.check_shell_commands()

	def check_shell_commands(self):
		"""Check for shell commands in profiles being imported.
		If there are any shell commands found, displays warning page
		and lets user to confirm import of them.

		Othewise, goes straight to next page as if user already confirmed them.
		"""
		grShellCommands = self.builder.get_object("grShellCommands")
		tvShellCommands = self.builder.get_object("tvShellCommands")
		files = self.builder.get_object("lstImportPackage")
		model = tvShellCommands.get_model()
		model.clear()
		# Get all shell commands in all profiles
		for trash, trash, trash, trash, obj in files:
			if isinstance(obj.obj, Profile):
				for a in obj.obj.get_all_actions():
					if isinstance(a, ShellCommandAction):
						model.append((False, a.command))

		if len(model) > 0:
			# If there is shell command present, jump to warning page
			self.next_page(grShellCommands)
			btNext = self.enable_next(True, self.shell_import_confirmed)
			btNext.set_label(_("Continue"))
			btNext.set_sensitive(False)
		else:
			# Otherwise continue to next one
			self.shell_import_confirmed()

	def on_crShellCommandChecked_toggled(self, cr, path):
		tvShellCommands = self.builder.get_object("tvShellCommands")
		btNext = self.builder.get_object("btNext")
		model = tvShellCommands.get_model()
		model[path][0] = not model[path][0]
		btNext.set_sensitive(True)
		for row in model:
			if not row[0]:
				btNext.set_sensitive(False)
				return

	def shell_import_confirmed(self):
		grSccImportFinished = self.builder.get_object("grSccImportFinished")
		lblSccImportFinished = self.builder.get_object("lblSccImportFinished")
		txName2 = self.builder.get_object("txName2")
		files = self.builder.get_object("lstImportPackage")
		vbImportPackage = self.builder.get_object("vbImportPackage")

		enabled, trash, name, trash, obj = files[0]
		lblSccImportFinished.set_text(_("Profile sucessfully imported"))
		txName2.set_text(name)
		vbImportPackage.set_visible(len(files) > 1)
		self.next_page(grSccImportFinished)
		self.on_txName2_changed()

	def on_txName2_changed(self, *a):
		txName2 = self.builder.get_object("txName2")
		btNext = self.enable_next(True, self.on_scc_import_confirmed)
		files = self.builder.get_object("lstImportPackage")
		cbImportHidden = self.builder.get_object("cbImportPackageHidden")
		cbImportVisible = self.builder.get_object("cbImportPackageVisible")
		cbImportNone = self.builder.get_object("cbImportPackageNone")
		rvAdvanced = self.builder.get_object("rvImportPackageAdvanced")
		btNext.set_label("Apply")
		btNext.set_use_stock(True)
		main_name = txName2.get_text()
		if self.check_name(main_name):
			btNext.set_sensitive(True)
		else:
			btNext.set_sensitive(False)

		cbImportHidden.set_label(_('Import as hidden menus and profiles named ".%s:name"') % (main_name,))
		cbImportVisible.set_label(_('Import normaly, with names formated as "%s:name"') % (main_name,))

		for i in range(len(files)):
			enabled, name, importas, type, obj = files[i]
			if enabled == 2:
				importas = main_name
			elif cbImportHidden.get_active():
				importas = ".%s:%s" % (main_name, name)
				enabled = 1
			elif cbImportVisible.get_active():
				importas = "%s:%s" % (main_name, name)
				enabled = 1
			elif cbImportNone.get_active():
				enabled = 0
			files[i] = enabled, name, importas, type, obj

	def on_cbImportPackageAdvanced_toggled(self, *a):
		rvImportPackageAdvanced = self.builder.get_object("rvImportPackageAdvanced")
		cbImportPackageAdvanced = self.builder.get_object("cbImportPackageAdvanced")
		rvImportPackageAdvanced.set_reveal_child(cbImportPackageAdvanced.get_active())

	def on_crIPKGEnabled_toggled(self, renderer, path):
		files = self.builder.get_object("lstImportPackage")
		i = int(path)
		enabled, name, importas, type, obj = files[i]
		# 1st rown cannot be toggled
		if enabled != 2:
			enabled = 1 if enabled == 0 else 0
			files[i] = enabled, name, importas, type, obj

	def on_crIPKGImportAs_edited(self, renderer, path, new_name):
		files = self.builder.get_object("lstImportPackage")
		txName2 = self.builder.get_object("txName2")
		i = int(path)
		enabled, name, importas, type, obj = files[i]
		importas = new_name
		if enabled == 2:
			txName2.set_text(importas)
		files[i] = enabled, name, importas, type, obj

	def on_scc_import_confirmed(self, *a):
		files = self.builder.get_object("lstImportPackage")
		new_profile_names = {}
		new_menu_names = {}
		for enabled, name, importas, trash, obj in files:
			if enabled != 0:
				if isinstance(obj.obj, Profile):
					new_profile_names[name] = importas
				elif isinstance(obj.obj, MenuData):
					new_menu_names["%s.menu" % (name,)] = "%s.menu" % (importas,)

		def apply_replacements(obj):
			for a in obj.get_all_actions():
				if isinstance(a, ChangeProfileAction):
					if a.profile in new_profile_names:
						a.profile = new_profile_names[a.profile]
						a.parameters = tuple([a.profile] + list(a.parameters[1:]))
				elif isinstance(a, MenuAction):
					if a.menu_id in new_menu_names:
						a.menu_id = new_menu_names[a.menu_id]
						a.parameters = tuple([a.menu_id] + list(a.parameters[1:]))

		for enabled, trash, importas, trash, obj in files:
			if enabled != 0:
				# TODO: update references
				if isinstance(obj.obj, Profile):
					apply_replacements(obj.obj)
					obj.obj.save(os.path.join(get_profiles_path(), "%s.sccprofile" % (importas,)))
				elif isinstance(obj.obj, MenuData):
					apply_replacements(obj.obj)
					jstr = Encoder(sort_keys=True, indent=4).encode(obj.obj)
					filename = os.path.join(get_menus_path(), "%s.menu" % (importas,))
					open(filename, "w").write(jstr)

		trash, trash, importas, trash, obj = files[0]  # 1st is always profile that's being imported
		self.app.new_profile(obj.obj, importas)
		GLib.idle_add(self.window.destroy)



#!/usr/bin/env python3
"""SC-Controller - Input Display
"""


import logging
import os
import signal
import sys

from gi.repository import Gtk

from scc.constants import LEFT, RIGHT, STICK, STICK_PAD_MAX, SCButtons
from scc.gui.daemon_manager import DaemonManager
from scc.gui.svg_widget import SVGWidget
from scc.osd import OSDWindow

log = logging.getLogger("osd.InputDisplay")


class InputDisplay(OSDWindow):
	IMAGE = "inputdisplay.svg"
	HILIGHT_COLOR = "#FF00FF00"  # ARGB
	OBSERVE_COLOR = "#00007FFF"  # ARGB

	def __init__(self, imagepath="/usr/share/scc/images"):
		OSDWindow.__init__(self, "osd-menu")
		self.daemon = None
		self.config = None
		self.hilights = {self.HILIGHT_COLOR: set(), self.OBSERVE_COLOR: set()}
		self.imagepath = imagepath

		self._eh_ids = []

	def show(self):
		self.main_area = Gtk.Fixed()
		self.background = SVGWidget(os.path.join(self.imagepath, self.IMAGE))
		self.lpadTest = Gtk.Image.new_from_file(os.path.join(self.imagepath, "inputdisplay-cursor.svg"))
		self.rpadTest = Gtk.Image.new_from_file(os.path.join(self.imagepath, "inputdisplay-cursor.svg"))
		self.stickTest = Gtk.Image.new_from_file(os.path.join(self.imagepath, "inputdisplay-cursor.svg"))

		self.main_area.set_property("margin-left", 10)
		self.main_area.set_property("margin-right", 10)
		self.main_area.set_property("margin-top", 10)
		self.main_area.set_property("margin-bottom", 10)

		self.main_area.put(self.background, 0, 0)
		self.main_area.put(self.lpadTest, 40, 40)
		self.main_area.put(self.rpadTest, 290, 90)
		self.main_area.put(self.stickTest, 150, 40)

		self.add(self.main_area)

		OSDWindow.show(self)
		self.lpadTest.hide()
		self.rpadTest.hide()
		self.stickTest.hide()

	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process.
		use_config() should be be called before parse_argumets() if this is used.
		"""
		self.daemon = d
		self._connect_handlers()
		self.on_daemon_connected(self.daemon)

	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def on_daemon_connected(self, *a):
		c = self.daemon.get_controllers()[0]
		c.unlock_all()
		c.observe(
			DaemonManager.nocallback,
			self.on_observe_failed,
			"A",
			"B",
			"C",
			"X",
			"Y",
			"START",
			"BACK",
			"LB",
			"RB",
			"LPAD",
			"RPAD",
			"LGRIP",
			"RGRIP",
			"LT",
			"RT",
			"LEFT",
			"RIGHT",
			"STICK",
			"STICKPRESS",
		)
		c.connect("event", self.on_daemon_event_observer)
		c.connect("lost", self.on_controller_lost)

	def on_observe_failed(self, error):
		log.error("Failed to enable test mode: %s", error)
		if "Sniffing" in error:
			log.error("")
			log.error("=================================================================================")
			log.error("[!!] Please, enable 'Input Test Mode' on 'Advanced' tab in SC Controller settings")
			log.error("=================================================================================")
		self.quit(3)

	def on_daemon_event_observer(self, daemon, what, data):
		if what in (LEFT, RIGHT, STICK):
			widget, area = {
				LEFT: (self.lpadTest, "LPADTEST"),
				RIGHT: (self.rpadTest, "RPADTEST"),
				STICK: (self.stickTest, "STICKTEST"),
			}[what]
			# Check if stick or pad is released
			if data[0] == data[1] == 0:
				widget.hide()
				return
			if not widget.is_visible():
				widget.show()
			# Grab values
			ax, ay, aw, trash = self.background.get_area_position(area)
			cw = widget.get_allocation().width
			# Compute center
			x, y = ax + aw * 0.5 - cw * 0.5, ay + 1.0 - cw * 0.5
			# Add pad position
			x += data[0] * aw / STICK_PAD_MAX * 0.5
			y -= data[1] * aw / STICK_PAD_MAX * 0.5
			# Move circle
			self.main_area.move(widget, x, y)
		elif what in ("LT", "RT", "STICKPRESS"):
			what = {"LT": "LEFT", "RT": "RIGHT", "STICKPRESS": "STICK"}[what]
			if data[0]:
				self.hilights[self.OBSERVE_COLOR].add(what)
			else:
				self.hilights[self.OBSERVE_COLOR].remove(what)
			self._update_background()
		elif hasattr(SCButtons, what):
			try:
				if data[0]:
					self.hilights[self.OBSERVE_COLOR].add(what)
				else:
					self.hilights[self.OBSERVE_COLOR].remove(what)
				self._update_background()
			except KeyError:
				# Non fatal
				pass
		else:
			print("event", what)

	def _update_background(self):
		h = {}
		for color in self.hilights:
			for i in self.hilights[color]:
				h[i] = color
		self.background.hilight(h)


def sigint(*a):
	print("\n*break*")
	sys.exit(0)


if __name__ == "__main__":
	signal.signal(signal.SIGINT, sigint)

	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.tools import init_logging

	init_logging()

	m = InputDisplay()
	if not m.parse_argumets(sys.argv):
		sys.exit(1)
	m.run()
	sys.exit(m.get_exit_code())



"""SC-Controller.

Copyright (C) 2018 Kozec

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as published by
the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import subprocess

# AppImageBuilder relies on its exec hooks to resolve relative shebang and ELF
# interpreter paths. Python 3.13+ uses posix_spawn() in more situations, which
# bypasses that compatibility path.
if os.environ.get("APPDIR") and hasattr(subprocess, "_USE_POSIX_SPAWN"):
	subprocess._USE_POSIX_SPAWN = False



"""SC Controller - Fake controller driver.

This driver does nothing by default, unless SCC_FAKES environment variable is
set. If it is, creates as many fake controller devices as integer stored in
SCC_FAKES says.

Created controllers are completely useless. For debuging purposes only.
"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

from scc.controller import Controller

if TYPE_CHECKING:
	from scc.sccdaemon import SCCDaemon

ENV_VAR = "SCC_FAKES"

if ENV_VAR in os.environ:
	log = logging.getLogger("FakeDrv")

	def init(daemon: SCCDaemon, config: dict) -> bool:
		return True

	def start(daemon: SCCDaemon) -> None:
		num = int(os.environ[ENV_VAR])
		log.debug("Creating %s fake controllers", num)
		for x in range(num):
			daemon.add_controller(FakeController(x))


class FakeController(Controller):
	def __init__(self, number: int) -> None:
		Controller.__init__(self)
		self._number = number
		self._id = f"fake{self._number}"

	def get_type(self) -> str:
		return "fake"

	def set_led_level(self, level: int) -> None:
		log.debug("FakeController %s led level set to %s", self.get_id(), level)

	def __repr__(self) -> str:
		return f"<FakeController {self.get_id()}>"



"""SC-Controller - OSD Menu.

Display menu that user can navigate through and print chosen item id to stdout
"""

import logging
import os
import sys
from xml.etree import ElementTree as ET

from gi.repository import Gdk, GdkPixbuf, GdkX11, Gtk

import scc.osd.osk_actions
from scc.actions import Action
from scc.config import Config
from scc.constants import (
	CPAD,
	LEFT,
	RIGHT,
	STICK,
	STICK_PAD_MAX,
	ControllerFlags,
	SCButtons,
)
from scc.gui.daemon_manager import DaemonManager
from scc.gui.keycode_to_key import KEY_TO_KEYCODE
from scc.gui.svg_widget import SVGEditor, SVGWidget
from scc.lib import xwrappers as X
from scc.modifiers import ModeModifier
from scc.osd import OSDWindow
from scc.osd.slave_mapper import SlaveMapper
from scc.osd.timermanager import TimerManager
from scc.parser import TalkingActionParser
from scc.paths import get_config_path, get_share_path
from scc.profile import Profile
from scc.tools import circle_to_square, clamp, find_button_image, find_profile
from scc.uinput import Keys

log = logging.getLogger("osd.keyboard")

SPECIAL_KEYS = {
	# Maps keycode to unicode character representing some
	# very special keys
	8: "←",
	9: "⇥",
	13: "↲",
	27: "␛",
	32: "␣",
}


class KeyboardImage(Gtk.DrawingArea):
	LINE_WIDTH = 2

	__gsignals__ = {}

	def __init__(self, image):
		Gtk.DrawingArea.__init__(self)
		self.connect("size-allocate", self.on_size_allocate)
		self.connect("draw", self.on_draw)

		areas = []
		self.color_button1 = 0.8, 0, 0, 1  # Just random mess,
		self.color_button1_border = 1, 0, 0, 1  # config overrides it anyway
		self.color_button2 = 0.8, 0.8, 0, 1
		self.color_button2_border = 1, 1, 0, 1
		self.color_hilight = 0, 1, 1, 1
		self.color_pressed = 1, 1, 1, 1
		self.color_text = 1, 1, 1, 1

		self.overlay = SVGWidget(image, False)
		self.tree = ET.fromstring(self.overlay.current_svg.encode("utf-8"))
		SVGWidget.find_areas(self.tree, None, areas, get_colors=True)

		self._hilight = ()
		self._pressed = ()
		self._button_images = {}
		self._help_areas = [self.get_limit("HELP_LEFT"), self.get_limit("HELP_RIGHT")]
		self._help_lines = ([], [])

		# TODO: It would be cool to use user-set font here, but cairo doesn't
		# have glyph replacement and most of default fonts (Ubuntu, Cantarell,
		# similar shit) misses pretty-much everything but letters, notably ↲
		#
		# For that reason, DejaVu Sans is hardcoded for now. On systems
		# where DejaVu Sans is not available, Cairo will automatically fallback
		# to default font.
		self.font_face = "DejaVu Sans"
		# self.font_face = Gtk.Label(label="X").get_style().font_desc.get_family()
		log.debug("Using font %s", self.font_face)

		self.buttons = [Button(self.tree, area) for area in areas]
		background = SVGEditor.find_by_id(self.tree, "BACKGROUND")
		self.set_size_request(*SVGEditor.get_size(background))
		self.overlay.edit().keep("overlay").commit()
		self.overlay.hilight({})
		# with open("/tmp/a.svg", "w") as file:
		# 	file.write(self.overlay.current_svg.encode("utf-8"))

	def hilight(self, hilight, pressed):
		self._hilight = hilight
		self._pressed = pressed
		self.queue_draw()

	def set_help(self, left, right):
		self._help_lines = (left, right)
		self.queue_draw()

	def set_labels(self, labels):
		for b in self.buttons:
			label = labels.get(b)
			if type(label) in (int,):
				pass
			elif label:
				# b.label = label.encode("utf-8")
				b.label = label
		self.queue_draw()

	def get_limit(self, id):
		a = SVGEditor.find_by_id(self.tree, id)
		width, height = 0, 0
		# if not hasattr(a, "parent"): a.parent = None

		x, y = SVGEditor.get_translation(a, absolute=True)
		if "width" in a.attrib:
			width = float(a.attrib["width"])
		if "height" in a.attrib:
			height = float(a.attrib["height"])
		return x, y, width, height

	@staticmethod
	def increase_contrast(buf):
		"""Takes input image, which is assumed to be grayscale RGBA and turns it
		into "symbolic" image by inverting colors of pixels where opacity is
		greater than threshold.
		"""
		pixels = [x for x in buf.get_pixels()]
		bpp = 4 if buf.get_has_alpha() else 3
		w, h = buf.get_width(), buf.get_height()
		stride = buf.get_rowstride()
		for i in range(0, len(pixels), bpp):
			if pixels[i + 3] > 64:
				pixels[i + 0] = 255 - pixels[i + 0]
				pixels[i + 1] = 255 - pixels[i + 1]
				pixels[i + 2] = 255 - pixels[i + 2]

		pixels = b"".join([chr(x).encode("latin-1") for x in pixels])
		rv = GdkPixbuf.Pixbuf.new_from_data(
			pixels, buf.get_colorspace(), buf.get_has_alpha(), buf.get_bits_per_sample(), w, h, stride, None,
		)
		rv.pixels = pixels  # Has to be kept in memory
		return rv

	def get_button_image(self, x, size):
		"""Loads and returns button image as pixbuf.
		Pixbufs are cached.
		"""
		if x not in self._button_images:
			path, bw = find_button_image(x, prefer_bw=True)
			if path is None:
				self._button_images[x] = None
				return None
			buf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, size, size)
			buf = self.increase_contrast(buf)
			self._button_images[x] = buf
		i = self._button_images[x]
		return i

	def on_draw(self, self2, ctx):
		ctx.select_font_face(self.font_face, 0, 0)

		ctx.set_line_width(self.LINE_WIDTH)
		ctx.set_font_size(48)
		ascent, descent, height, max_x_advance, max_y_advance = ctx.font_extents()

		# Buttons
		for button in self.buttons:
			if button in self._pressed:
				ctx.set_source_rgba(*self.color_pressed)
			elif button in self._hilight:
				ctx.set_source_rgba(*self.color_hilight)
			elif button.dark:
				ctx.set_source_rgba(*self.color_button2)
			else:
				ctx.set_source_rgba(*self.color_button1)
			# filled rectangle
			x, y, w, h = button
			ctx.move_to(x, y)
			ctx.line_to(x + w, y)
			ctx.line_to(x + w, y + h)
			ctx.line_to(x, y + h)
			ctx.line_to(x, y)
			ctx.fill()

			# border
			ctx.set_source_rgba(*self.color_button1_border)
			ctx.move_to(x, y)
			ctx.line_to(x + w, y)
			ctx.line_to(x + w, y + h)
			ctx.line_to(x, y + h)
			ctx.line_to(x, y)
			ctx.stroke()

			# label
			if button.label:
				ctx.set_source_rgba(*self.color_text)
				extents = ctx.text_extents(button.label)
				x_bearing, y_bearing, width, trash, x_advance, y_advance = extents
				ctx.move_to(x + w * 0.5 - width * 0.5 - x_bearing, y + h * 0.5 + height * 0.3)
				ctx.show_text(button.label)
				ctx.stroke()

		# Overlay
		Gdk.cairo_set_source_pixbuf(ctx, self.overlay.get_pixbuf(), 0, 0)
		ctx.paint()

		# Help
		ctx.set_source_rgba(*self.color_text)
		ctx.set_font_size(16)
		ascent, descent, height, max_x_advance, max_y_advance = ctx.font_extents()
		for left_right in (0, 1):
			x, y, w, h = self._help_areas[left_right]
			lines = self._help_lines[left_right]
			xx = x if left_right == 0 else x + w
			yy = y
			for icon, line in lines:
				yy += height
				if yy > y + h:
					break
				image = self.get_button_image(icon, height * 0.9)
				if image is None:
					continue
				iw, ih = image.get_width(), image.get_height()

				if left_right == 1:  # Right align
					extents = ctx.text_extents(line)
					x_bearing, y_bearing, width, trash, x_advance, y_advance = extents
					ctx.save()
					ctx.translate(xx - height + (height - iw) * 0.5, 1 + yy - (ascent + ih) * 0.5)
					Gdk.cairo_set_source_pixbuf(ctx, image, 0, 0)
					ctx.paint()
					ctx.restore()
					ctx.move_to(xx - x_bearing - width - 5 - height, yy)
				else:
					ctx.save()
					ctx.translate(1 + xx + (height - iw) * 0.5, 1 + yy - (ascent + ih) * 0.5)
					Gdk.cairo_set_source_pixbuf(ctx, image, 0, 0)
					ctx.paint()
					ctx.restore()
					ctx.move_to(xx + 5 + height, yy)

				ctx.show_text(line)
				ctx.stroke()

	def on_size_allocate(self, *a):
		pass


class Button:
	def __init__(self, tree, area):
		self.contains = area.contains
		self.name = area.name
		self.label = None
		self.x, self.y = area.x, area.y
		self.w, self.h = area.w, area.h
		self.dark = area.color[2] < 0.5  # Dark button is less than 50% blue

	def __iter__(self):
		return iter((self.x, self.y, self.w, self.h))


class Keyboard(OSDWindow, TimerManager):
	EPILOG = """Exit codes:
   0  - clean exit, user closed keyboard
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while keyboard is displayed.
   3  - erorr, failed to lock input stick, pad or button(s)
	"""
	OSK_PROF_NAME = ".scc-osd.keyboard"

	BUTTON_MAP = {
		SCButtons.A.name: Keys.KEY_ENTER,
		SCButtons.B.name: Keys.KEY_ESC,
		SCButtons.LB.name: Keys.KEY_BACKSPACE,
		SCButtons.RB.name: Keys.KEY_SPACE,
		SCButtons.LGRIP.name: Keys.KEY_LEFTSHIFT,
		SCButtons.RGRIP.name: Keys.KEY_RIGHTALT,
	}

	def __init__(self, config=None):
		self.kbimage = os.path.join(get_config_path(), "keyboard.svg")
		if not os.path.exists(self.kbimage):
			# Prefer image in ~/.config/scc, but load default one as fallback
			self.kbimage = os.path.join(get_share_path(), "images", "keyboard.svg")

		TimerManager.__init__(self)
		OSDWindow.__init__(self, "osd-keyboard")
		self.daemon = None
		self.mapper = None
		self.keymap = Gdk.Keymap.get_default()
		self.keymap.connect("state-changed", self.on_keymap_state_changed)
		Action.register_all(sys.modules["scc.osd.osk_actions"], prefix="OSK")
		self.profile = Profile(TalkingActionParser())
		self.config = config or Config()
		if isinstance(Gdk.Display.get_default(), GdkX11.X11Display):
			self.x11_dpy = X.Display(hash(GdkX11.x11_get_default_xdisplay()))
			self.group = None
		else:
			self.x11_dpy = None
			self.group = 0
		self.limits = {}
		self.background = None

		cursor = os.path.join(get_share_path(), "images", "menu-cursor.svg")
		self.cursors = {}
		self.cursors[LEFT] = Gtk.Image.new_from_file(cursor)
		self.cursors[LEFT].set_name("osd-keyboard-cursor")
		self.cursors[RIGHT] = Gtk.Image.new_from_file(cursor)
		self.cursors[RIGHT].set_name("osd-keyboard-cursor")
		self.cursors[CPAD] = Gtk.Image.new_from_file(cursor)
		self.cursors[CPAD].set_name("osd-keyboard-cursor")

		self._eh_ids = []
		self._controller = None
		self._stick = 0, 0
		self._hovers = {self.cursors[LEFT]: None, self.cursors[RIGHT]: None}
		self._pressed = {self.cursors[LEFT]: None, self.cursors[RIGHT]: None}
		self._pressed_areas = {}

		self.c = Gtk.Box()
		self.c.set_name("osd-keyboard-container")

		self.f = Gtk.Fixed()

	def _create_background(self):
		self.background = KeyboardImage(self.args.image)
		self.recolor()

		self.limits = {}
		self.limits[LEFT] = self.background.get_limit("LIMIT_LEFT")
		self.limits[RIGHT] = self.background.get_limit("LIMIT_RIGHT")
		self.limits[CPAD] = self.background.get_limit("LIMIT_CPAD")
		self._pack()

	def _pack(self):
		self.f.add(self.background)
		self.f.add(self.cursors[LEFT])
		self.f.add(self.cursors[RIGHT])
		self.f.add(self.cursors[CPAD])
		self.c.add(self.f)
		self.add(self.c)

	def recolor(self):
		# TODO: keyboard description is probably not needed anymore
		_get = lambda a: SVGWidget.color_to_float(self.config["osk_colors"].get(a, ""))
		self.background.color_button1 = _get("button1")
		self.background.color_button1_border = _get("button1_border")
		self.background.color_button2 = _get("button2")
		self.background.color_button2_border = _get("button2_border")
		self.background.color_hilight = _get("hilight")
		self.background.color_pressed = _get("pressed")
		self.background.color_text = _get("text")

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process
		"""
		self.daemon = d
		self._cononect_handlers()
		self.on_daemon_connected(self.daemon)

	def redraw_background(self, *a) -> None:
		"""Forces a repaint of the keyboard background image.

		Called by the OSD daemon after recolor()/update_labels() when the
		OSD color configuration changes while the keyboard is visible.
		"""
		if self.background is not None:
			self.background.queue_draw()

	def on_keymap_state_changed(self, x11keymap):
		if not self.timer_active("labels"):
			self.timer("labels", 0.1, self.update_labels)

	def set_help(self):
		"""Updates help shown on keyboard image.
		Keyboard bindings don't change on the fly, so this is done only
		right after start or when daemon is reconfigured.
		"""
		if self._controller is None:
			# Not yet connected
			return
		gui_config = self._controller.load_gui_config(os.path.join(get_share_path(), "images"))
		l_lines, r_lines, used = [], [], set()

		def add_action(side, button, a):
			if not a:
				return
			if isinstance(a, scc.osd.osk_actions.OSKCursorAction):
				if a.side != CPAD:
					return
			if isinstance(a, ModeModifier):
				for x in a.get_child_actions():
					add_action(side, button, x)
				return
			desc = a.describe(Action.AC_OSK)
			if desc in used:
				if isinstance(a, scc.osd.osk_actions.OSKPressAction):
					# Special case, both triggers are set to "press a key"
					pass
				else:
					return
			icon = self._controller.get_button_name(gui_config, button)
			side.append((icon, desc))
			used.add(desc)

		def add_button(side, b):
			add_action(side, b, self.profile.buttons[b])

		if self._controller.get_flags() & ControllerFlags.NO_GRIPS == 0:
			add_button(l_lines, SCButtons.LGRIP)
			add_button(r_lines, SCButtons.RGRIP)
		add_action(l_lines, SCButtons.LT, self.profile.triggers[LEFT])
		add_action(r_lines, SCButtons.RT, self.profile.triggers[RIGHT])
		for b in (SCButtons.LB, SCButtons.Y, SCButtons.X):
			add_button(l_lines, b)
		for b in (SCButtons.RB, SCButtons.B, SCButtons.A):
			add_button(r_lines, b)

		if self._controller.get_flags() & ControllerFlags.HAS_CPAD != 0:
			for lst in (l_lines, r_lines):
				while len(lst) > 3:
					lst.pop()
				while len(lst) < 3:
					lst.append((None, ""))
			add_action(r_lines, CPAD, self.profile.pads[CPAD])
		add_action(l_lines, SCButtons.STICKPRESS, self.profile.stick)

		self.background.set_help(l_lines, r_lines)

	def update_labels(self):
		"""Updates keyboard labels based on active X keymap"""
		labels = {}
		# Get current layout group
		if self.x11_dpy is not None:
			self.group = X.get_xkb_state(self.x11_dpy).group
		# Get state of shift/alt/ctrl key
		mt = Gdk.ModifierType(self.keymap.get_modifier_state())
		for button in self.background.buttons:
			if getattr(Keys, button.name, None) in KEY_TO_KEYCODE:
				keycode = KEY_TO_KEYCODE[getattr(Keys, button.name)]
				translation = self.keymap.translate_keyboard_state(keycode, mt, self.group)
				if hasattr(translation, "keyval"):
					code = Gdk.keyval_to_unicode(translation.keyval)
				else:
					code = Gdk.keyval_to_unicode(translation[1])
				if code >= 33:  # Printable chars, w/out space
					labels[button] = chr(code).strip()
				else:
					labels[button] = SPECIAL_KEYS.get(code)
		self.background.set_labels(labels)

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument("image", type=str, nargs="?", default=self.kbimage, help="keyboard image to use")

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		return True

	def _cononect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("reconfigured", self.on_reconfigured)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._cononect_handlers()
		OSDWindow.run(self)

	def load_profile(self):
		self.profile.load(find_profile(Keyboard.OSK_PROF_NAME)).compress()
		self.set_help()

	def on_reconfigured(self, *a):
		self.load_profile()
		log.debug("Reloaded profile")

	def on_daemon_connected(self, *a):
		def success(*a):
			log.info("Sucessfully locked input")

		c = self.choose_controller(self.daemon)
		if c is None or not c.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		self._eh_ids += [
			(c, c.connect("event", self.on_event)),
			(c, c.connect("lost", self.on_controller_lost)),
		]

		# TODO: Single-handed mode for PS4 posponed
		locks = [LEFT, RIGHT, STICK, "STICKPRESS"] + [b.name for b in SCButtons]
		if (c.get_flags() & ControllerFlags.HAS_CPAD) == 0:
			# Two pads, two hands
			locks = [LEFT, RIGHT, STICK, "STICKPRESS"] + [b.name for b in SCButtons]
			self.cursors[CPAD].hide()
		else:
			# Single-handed mode
			locks = [CPAD, "CPADPRESS", STICK, "STICKPRESS"] + [b.name for b in SCButtons]
			self._hovers[self.cursors[RIGHT]] = None
			self._hovers = {self.cursors[CPAD]: None}
			self._pressed = {self.cursors[CPAD]: None}
			self.cursors[LEFT].hide()
			self.cursors[RIGHT].hide()

			# There is no configurable nor default mapping for CPDAD,
			# so situable mappings are hardcoded here
			self.profile.pads[CPAD] = scc.osd.osk_actions.OSKCursorAction(CPAD)
			self.profile.pads[CPAD].speed = [0.85, 1.2]
			self.profile.buttons[SCButtons.CPADPRESS] = scc.osd.osk_actions.OSKPressAction(CPAD)

			for i in (LEFT, RIGHT):
				if isinstance(self.profile.triggers[i], scc.osd.osk_actions.OSKPressAction):
					self.profile.triggers[i] = scc.osd.osk_actions.OSKPressAction(CPAD)

		self._controller = c
		c.lock(success, self.on_failed_to_lock, *locks)
		self.set_help()

	def quit(self, code=-1):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		del self.mapper
		OSDWindow.quit(self, code)

	def show(self, *a):
		if self.background is None:
			self._create_background()
		OSDWindow.show(self, *a)
		self.load_profile()
		self.mapper = SlaveMapper(self.profile, None, keyboard=b"SCC OSD Keyboard", mouse=b"SCC OSD Mouse")
		self.mapper.set_special_actions_handler(self)
		self.set_cursor_position(0, 0, self.cursors[LEFT], self.limits[LEFT])
		self.set_cursor_position(0, 0, self.cursors[RIGHT], self.limits[RIGHT])
		self.set_cursor_position(0, 0, self.cursors[CPAD], self.limits[CPAD])
		self.timer("labels", 0.1, self.update_labels)

	def on_event(self, daemon, what, data):
		"""Called when button press, button release or stick / pad update is sent by daemon."""
		if self.x11_dpy is not None:
			group = X.get_xkb_state(self.x11_dpy).group
			if self.group != group:
				self.group = group
				self.timer("labels", 0.1, self.update_labels)
		self.mapper.handle_event(daemon, what, data)

	def on_sa_close(self, *a):
		"""Called by CloseOSDKeyboardAction"""
		self.quit(0)

	def on_sa_cursor(self, mapper, action, x, y):
		self.set_cursor_position(
			x * action.speed[0], y * action.speed[1], self.cursors[action.side], self.limits[action.side],
		)

	def on_sa_move(self, mapper, action, x, y):
		self._stick = x, y
		if not self.timer_active("stick"):
			self.timer("stick", 0.05, self._move_window)

	def on_sa_press(self, mapper, action, pressed):
		self.key_from_cursor(self.cursors[action.side], pressed)

	def set_cursor_position(self, x, y, cursor, limit):
		"""Moves cursor image."""
		if cursor not in self._hovers:
			return
		w = limit[2] - (cursor.get_allocation().width * 0.5)
		h = limit[3] - (cursor.get_allocation().height * 0.5)
		x = x / float(STICK_PAD_MAX)
		y = y / float(STICK_PAD_MAX) * -1.0

		x, y = circle_to_square(x, y)

		x = clamp(
			cursor.get_allocation().width * 0.5,
			(limit[0] + w * 0.5) + x * w * 0.5,
			self.get_allocation().width - cursor.get_allocation().width,
		)

		y = clamp(
			cursor.get_allocation().height * 0.5,
			(limit[1] + h * 0.5) + y * h * 0.5,
			self.get_allocation().height - cursor.get_allocation().height,
		)

		cursor.position = int(x), int(y)
		self.f.move(cursor, x - cursor.get_allocation().width * 0.5, y - cursor.get_allocation().height * 0.5)
		for button in self.background.buttons:
			if button.contains(x, y):
				if button != self._hovers[cursor]:
					self._hovers[cursor] = button
					if self._pressed[cursor] is not None:
						self.mapper.keyboard.releaseEvent([self._pressed[cursor]])
						self.key_from_cursor(cursor, True)
					if not self.timer_active("update"):
						self.timer("update", 0.01, self.update_background)
					break

	def update_background(self, *whatever):
		"""Updates hilighted keys on bacgkround image.
		"""
		self.background.hilight(
			set([a for a in self._hovers.values() if a]), set([a for a in self._pressed_areas.values() if a]),
		)

	def _move_window(self, *a):
		"""Called by timer while stick is tilted to move window around the screen.
		"""
		x, y = self._stick
		x = x * 50.0 / STICK_PAD_MAX
		y = y * -50.0 / STICK_PAD_MAX
		rx, ry = self.get_position()
		self.move(rx + x, ry + y)
		if abs(self._stick[0]) > 100 or abs(self._stick[1]) > 100:
			self.timer("stick", 0.05, self._move_window)

	def key_from_cursor(self, cursor, pressed):
		"""Sends keypress/keyrelease event to emulated keyboard, based on
		position of cursor on OSD keyboard.
		"""
		x, y = cursor.position

		if pressed:
			for button in self.background.buttons:
				if button.contains(x, y):
					if button.name.startswith("KEY_") and hasattr(Keys, button.name):
						key = getattr(Keys, button.name)
						if self._pressed[cursor] is not None:
							self.mapper.keyboard.releaseEvent([self._pressed[cursor]])
						self.mapper.keyboard.pressEvent([key])
						self._pressed[cursor] = key
						self._pressed_areas[cursor] = button
					break
		elif self._pressed[cursor] is not None:
			self.mapper.keyboard.releaseEvent([self._pressed[cursor]])
			self._pressed[cursor] = None
			del self._pressed_areas[cursor]
		if not self.timer_active("update"):
			self.timer("update", 0.01, self.update_background)


def main():
	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.tools import init_logging

	init_logging()

	k = Keyboard()
	if not k.parse_argumets(sys.argv):
		sys.exit(1)
	k.run()


if __name__ == "__main__":
	import signal

	def sigint(*a):
		print("\n*break*")
		sys.exit(-1)

	signal.signal(signal.SIGINT, sigint)
	main()



#!/usr/bin/env python3
import signal
import sys


def main() -> None:
	def sigint(*a):
		print("\n*break*")
		sys.exit(0)

	if __name__ == "__main__":
		signal.signal(signal.SIGINT, sigint)

		import gi

		gi.require_version("Gtk", "3.0")
		gi.require_version("Rsvg", "2.0")
		gi.require_version("GdkX11", "3.0")

		from scc.tools import init_logging

		init_logging()

		from scc.osd.message import Message

		m = Message()
		if not m.parse_argumets(sys.argv):
			sys.exit(1)
		m.run()
		sys.exit(m.get_exit_code())


if __name__ == "__main__":
	main()



#!/usr/bin/env python3
"""SC-Controller - Action Editor

Allows to edit button or trigger action.
"""


import logging
import os

from scc.actions import AxisAction, HatDownAction, HatLeftAction, HatRightAction, HatUpAction, MouseAction
from scc.gui.area_to_action import AREA_TO_ACTION
from scc.gui.editor import Editor
from scc.gui.svg_widget import SVGWidget

log = logging.getLogger("Chooser")

AXIS_ACTION_CLASSES = (AxisAction, MouseAction, HatLeftAction, HatRightAction, HatUpAction, HatDownAction)


class Chooser(Editor):
	IMAGES = {}

	ACTIVE_COLOR = "#FF00FF00"  # ARGB
	HILIGHT_COLOR = "#FFFF0000"  # ARGB

	def __init__(self, app):
		self.app = app
		self.active_area = None  # Area that is permanently hilighted on the image
		self.images = []
		self.axes_allowed = True
		self.mouse_allowed = True

	def setup_image(self, grid_columns=0):
		for id in self.IMAGES:
			parent = self.builder.get_object(id)
			if parent is not None:
				image = SVGWidget(os.path.join(self.app.imagepath, self.IMAGES[id]))
				image.connect("hover", self.on_background_area_hover)
				image.connect("leave", self.on_background_area_hover, None)
				image.connect("click", self.on_background_area_click)
				self.images.append(image)
				if grid_columns:
					# Grid
					parent.attach(image, 0, 0, grid_columns, 1)
				else:
					# Box
					parent.pack_start(image, True, True, 0)
				parent.show_all()

	def set_active_area(self, a):
		"""Sets area that is permanently hilighted on image.
		"""
		self.active_area = a
		for i in self.images:
			i.hilight({self.active_area: Chooser.ACTIVE_COLOR})

	def on_background_area_hover(self, background, area):
		if area in AREA_TO_ACTION:
			if AREA_TO_ACTION[area][0] in AXIS_ACTION_CLASSES:
				if not self.axes_allowed:
					return
			if not self.mouse_allowed and "MOUSE" in area:
				return
		background.hilight({self.active_area: Chooser.ACTIVE_COLOR, area: Chooser.HILIGHT_COLOR})

	def on_background_area_click(self, trash, area):
		"""Called when user clicks on defined area on gamepad image.
		"""
		if area in AREA_TO_ACTION:
			cls, params = AREA_TO_ACTION[area][0], AREA_TO_ACTION[area][1:]
			if not self.axes_allowed and cls in AXIS_ACTION_CLASSES:
				return
			if not self.mouse_allowed and "MOUSE" in area:
				return
			self.area_action_selected(area, cls(*params))
		else:
			log.warning("Click on unknown area: %s" % (area,))

	def area_action_selected(self, area, action):
		raise Exception("Override me!")

	def hide_axes(self):
		"""Prevents user from selecting axes"""
		self.axes_allowed = False

	def hide_mouse(self):
		"""Prevents user from selecting mouse-related stuff"""
		self.mouse_allowed = False



#!/usr/bin/env python3
import os
import signal
import sys


def main() -> None:
	def sigint(*a):
		print("\n*break*")
		sys.exit(0)

	signal.signal(signal.SIGINT, sigint)

	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("GdkX11", "3.0")
	gi.require_version("Rsvg", "2.0")

	from scc.paths import get_share_path
	from scc.tools import init_logging

	init_logging()

	from gi.repository import Gtk

	glades = os.path.join(get_share_path(), "glade")
	images = os.path.join(get_share_path(), "images")
	if Gtk.IconTheme.get_default():
		Gtk.IconTheme.get_default().append_search_path(images)
	# GObject.threads_init()

	from scc.gui.app import App

	App(glades, images).run(sys.argv)


if __name__ == "__main__":
	main()



"""SC-Controller - OSD Menu Generators

Auto-generated menus with stuff like list of all available profiles...
"""

import logging
import os
import traceback
from ctypes import POINTER, cast

from gi.repository import Gdk, GdkX11, Gio

from scc.lib import xwrappers as X
from scc.menu_data import MENU_GENERATORS, MenuGenerator, MenuItem
from scc.paths import get_default_profiles_path, get_profiles_path
from scc.tools import _, find_profile

log = logging.getLogger("osd.menu_gen")


class ProfileListMenuGenerator(MenuGenerator):
	"""Generates list of all available profiles"""

	GENERATOR_NAME = "profiles"

	@staticmethod
	def callback(menu, daemon, controller, menuitem):
		controller.set_profile(menuitem.filename)
		menu.hide()

		def on_response(*a):
			menu.quit(-2)

		daemon.request(b"OSD: " + menuitem.label.encode("utf-8") + b"\n", on_response, on_response)

	def describe(self):
		return _("[ All Profiles ]")

	def generate(self, menuhandler):
		# TODO: Cannot load directory content asynchronously here and I'm
		# TODO: not happy about it
		rv, all_profiles = [], {}
		for d in (get_default_profiles_path(), get_profiles_path()):
			for x in os.listdir(d):
				if x.endswith(".sccprofile") and not x.startswith("."):
					all_profiles[x] = os.path.join(d, x)
		for p in sorted(all_profiles, key=lambda s: s.lower()):
			menuitem = MenuItem("generated", p[0:-11])  # strips ".sccprofile"
			menuitem.filename = all_profiles[p]
			menuitem.callback = self.callback
			rv.append(menuitem)
		return rv


class RecentListMenuGenerator(MenuGenerator):
	"""Generates list of X recently used profiles"""

	GENERATOR_NAME = "recent"

	def __init__(self, rows=5, **b):
		MenuGenerator.__init__(self)
		self.rows = rows

	def generate(self, menuhandler):
		return _("[ %s Recent Profiles ]") % (self.rows,)

	def encode(self):
		return {"generator": self.GENERATOR_NAME, "rows": self.rows}

	def callback(self, menu, daemon, controller, menuitem):
		controller.set_profile(menuitem.filename)
		menu.hide()

		def on_response(*a):
			menu.quit(-2)

		daemon.request(b"OSD: " + menuitem.label.encode("utf-8") + b"\n", on_response, on_response)

	def generate(self, menuhandler):
		rv = []
		for p in menuhandler.config["recent_profiles"]:
			filename = find_profile(p)
			if filename:
				menuitem = MenuItem("generated", p)
				menuitem.filename = filename
				menuitem.callback = ProfileListMenuGenerator.callback
				rv.append(menuitem)
			if len(rv) >= self.rows:
				break
		return rv


class WindowListMenuGenerator(MenuGenerator):
	"""Generates list of all windows"""

	GENERATOR_NAME = "windowlist"
	MAX_LENGHT = 50

	# def generate(self, menuhandler):
	# return _("[ Window Lists ]")

	def encode(self):
		return {"generator": self.GENERATOR_NAME}

	@staticmethod
	def callback(menu, daemon, controller, menuitem):
		try:
			xid = int(menuitem.id)
			display = Gdk.Display.get_default()
			window = GdkX11.X11Window.foreign_new_for_display(display, xid)
			window.focus(0)
		except Exception:
			log.error("Failed to activate window")
			log.error(traceback.format_exc())
		menu.quit(-2)

	def generate(self, menuhandler):
		rv = []
		dpy = X.Display(hash(GdkX11.x11_get_default_xdisplay()))  # Magic
		root = X.get_default_root_window(dpy)

		count, wlist = X.get_window_prop(dpy, root, b"_NET_CLIENT_LIST", 1024)
		skip_taskbar = X.intern_atom(dpy, b"_NET_WM_STATE_SKIP_TASKBAR", True)
		wlist = cast(wlist, POINTER(X.XID))[0:count]
		for win in wlist:
			if skip_taskbar not in X.get_wm_state(dpy, win):
				title = X.get_window_title(dpy, win)[0 : self.MAX_LENGHT]
				menuitem = MenuItem(str(win), title)
				menuitem.callback = WindowListMenuGenerator.callback
				rv.append(menuitem)
		return rv


class GameListMenuGenerator(MenuGenerator):
	"""Generates list of applications known to XDG menu
	and belonging to 'Game' category
	"""

	GENERATOR_NAME = "games"
	MAX_LENGHT = 50

	_games = None  # Static list of know games

	# def generate(self, menuhandler):
	# return _("[ Games ]")

	def encode(self):
		return {"generator": self.GENERATOR_NAME}

	@staticmethod
	def callback(menu, daemon, controller, menuitem):
		menuitem._desktop_file.launch()
		menu.quit(-2)

	def generate(self, menuhandler):
		if GameListMenuGenerator._games is None:
			GameListMenuGenerator._games = []
			id = 0
			for x in Gio.AppInfo.get_all():
				if x.get_categories():
					if "Game" in x.get_categories().split(";"):
						menuitem = MenuItem(str(id), x.get_display_name(), icon=x.get_icon())
						menuitem.callback = GameListMenuGenerator.callback
						menuitem._desktop_file = x
						GameListMenuGenerator._games.append(menuitem)
		return GameListMenuGenerator._games


def register_menu_generators() -> None:
	"""Add classes to MENU_GENERATORS dict

	Needs to be called from menu otherwise entries will be empty
	"""
	for cls in (
		ProfileListMenuGenerator,
		RecentListMenuGenerator,
		WindowListMenuGenerator,
		GameListMenuGenerator,
	):
		MENU_GENERATORS[cls.GENERATOR_NAME] = cls



"""Imports VDF profile and converts it to Profile object."""

import logging
from typing import Self

from scc.actions import (
	Action,
	AxisAction,
	ButtonAction,
	DPadAction,
	HatDownAction,
	HatLeftAction,
	HatRightAction,
	HatUpAction,
	MouseAction,
	MultiAction,
	NoAction,
	RelAreaAction,
	TriggerAction,
	XYAction,
)
from scc.constants import ROLL, TRIGGER_CLICK, YAW, HapticPos, SCButtons
from scc.lib.vdf import ensure_list, parse_vdf
from scc.menu_data import MenuData, MenuItem
from scc.modifiers import (
	BallModifier,
	DoubleclickModifier,
	FeedbackModifier,
	HoldModifier,
	ModeModifier,
	SensitivityModifier,
)
from scc.parser import ActionParser, ParseError
from scc.profile import Profile
from scc.special_actions import ChangeProfileAction, GridMenuAction, MenuAction, RadialMenuAction
from scc.uinput import Axes, Keys, Rels

log = logging.getLogger("import.vdf")


class VDFProfile(Profile):
	BUTTON_TO_BUTTON = {
		# maps button keys from vdf file to SCButtons constants
		# It looks like uppercase button and lowercase button aliases are needed
		"button_a": SCButtons.A,
		"button_A": SCButtons.A,
		"button_b": SCButtons.B,
		"button_B": SCButtons.B,
		"button_x": SCButtons.X,
		"button_X": SCButtons.X,
		"button_y": SCButtons.Y,
		"button_Y": SCButtons.Y,
		"button_back_left": SCButtons.LGRIP,
		"button_back_right": SCButtons.RGRIP,
		"button_menu": SCButtons.BACK,
		"button_escape": SCButtons.START,  # what what what
		"left_bumper": SCButtons.LB,
		"right_bumper": SCButtons.RB,
		"left_click": SCButtons.LPAD,
		"right_click": SCButtons.RPAD,
	}

	SPECIAL_KEYS = {
		# Maps some key names from vdf file to Keys.* constants.
		# Rest of key names are converted in convert_key_name.
		"FORWARD_SLASH": Keys.KEY_SLASH,
		"VOLUME_DOWN": Keys.KEY_VOLUMEDOWN,
		"VOLUME_UP": Keys.KEY_VOLUMEUP,
		"NEXT_TRACK": Keys.KEY_NEXTSONG,
		"PREV_TRACK": Keys.KEY_PREVIOUSSONG,
		"PAGE_UP": Keys.KEY_PAGEUP,
		"PAGE_DOWN": Keys.KEY_PAGEDOWN,
		"SINGLE_QUOTE": Keys.KEY_APOSTROPHE,
		"DASH": Keys.KEY_MINUS,
		"RETURN": Keys.KEY_ENTER,
		"ESCAPE": Keys.KEY_ESC,
		"PERIOD": Keys.KEY_DOT,
		"LEFT_BRACKET": Keys.KEY_LEFTBRACE,
		"RIGHT_BRACKET": Keys.KEY_RIGHTBRACE,
		"KEYPAD_DASH": Keys.KEY_KPMINUS,
		"KEYPAD_FORWARD_SLASH": Keys.KEY_KPSLASH,
		"LEFT_CONTROL": Keys.KEY_LEFTCTRL,
		"RIGHT_CONTROL": Keys.KEY_RIGHTCTRL,
	}

	SPECIAL_BUTTONS = {
		# As SPECIAL_KEYS, but for buttons.
		"shoulder_left": Keys.BTN_TL,
		"shoulder_right": Keys.BTN_TR,
		"joystick_left": Keys.BTN_THUMBL,
		"joystick_right": Keys.BTN_THUMBR,
	}

	REGION_IMPORT_FACTOR = 0.6  # Bulgarian const.

	def __init__(self, name="Unnamed"):
		Profile.__init__(self, ActionParser())
		self.name = name
		self.next_menu_id = 1
		self.action_set_id = 0
		self.action_sets = {"default": self}
		self.action_set_switches = set()

	def parse_action(self, lst_or_str, button=None):
		"""Parses action from vdf file. a_string can be either string or list of
		strings, in which case MultiAction is returned.

		Returns Action instance or ParseError if action is not recognized.
		"""
		if type(lst_or_str) == list:
			return MultiAction.make(*[self.parse_action(x) for x in lst_or_str])
		# Split string into binding type, name and parameters
		binding, params = lst_or_str.split(" ", 1)
		if "," in params:
			params, name = params.split(",", 1)
		else:
			params, name = params, None
		params = params.split(" ")
		if name:
			name = name.strip()
		# Return apropriate Action for binding type
		if binding in ("key_press", "mouse_button"):
			if binding == "mouse_button":
				b = VDFProfile.convert_button_name(params[0])
			else:
				b = VDFProfile.convert_key_name(params[0])
			return ButtonAction(b).set_name(name)
		if binding == "xinput_button":
			# Special cases, as dpad is apparently button on Windows
			b = params[0].strip().lower()
			if b == "dpad_up":
				return HatUpAction(Axes.ABS_HAT0Y)
			if b == "dpad_down":
				return HatDownAction(Axes.ABS_HAT0Y)
			if b == "dpad_left":
				return HatLeftAction(Axes.ABS_HAT0X)
			if b == "dpad_right":
				return HatRightAction(Axes.ABS_HAT0X)
			if b == "trigger_left":
				return AxisAction(Axes.ABS_Z)
			if b == "trigger_right":
				return AxisAction(Axes.ABS_RZ)
			b = VDFProfile.convert_button_name(b)
			return ButtonAction(b).set_name(name)
		if binding in ("mode_shift"):
			if button is None:
				log.warning("Ignoring modeshift assigned to no button: '%s'" % (lst_or_str,))
				return NoAction()
			if button not in VDFProfile.BUTTON_TO_BUTTON:
				log.warning("Ignoring modeshift assigned to unknown button: '%s'" % (button,))
				return NoAction()
			self.modeshift_buttons[VDFProfile.BUTTON_TO_BUTTON[button]] = (params[1], params[0])
			return NoAction()
		if binding in ("controller_action"):
			if params[0] == "CHANGE_PRESET":
				id = int(params[1]) - 1
				cpa = ChangeProfileAction("action_set:%s" % (id,))
				self.action_set_switches.add(cpa)
				return cpa

			log.warning("Ignoring controller_action '%s' binding" % (params[0],))
			return NoAction()
		if binding == "mouse_wheel":
			if params[0].lower() == "scroll_down":
				return MouseAction(Rels.REL_WHEEL, -1)
			return MouseAction(Rels.REL_WHEEL, 1)
		if binding == "game_action":
			log.warning("Ignoring game_action binding: '%s'" % (lst_or_str,))
			return NoAction()

		raise ParseError("Unknown binding: '%s'" % (binding,))

	@staticmethod
	def parse_modifiers(group, action, side):
		"""If passed group or activator has 'settings' key, converts known
		settings to one or more Modifier.

		Returns resulting Action
		"""
		if "settings" in group:
			settings = group["settings"]
			sens = 1.0, 1.0, 1.0
			if "sensitivity" in settings:
				s = float(settings["sensitivity"]) / 100.0
				sens = s, s, s
			if "haptic_intensity" in settings:
				action = FeedbackModifier(
					HapticPos.LEFT if side == Profile.LEFT else HapticPos.RIGHT,
					512 * int(settings["haptic_intensity"]),
					8,
					action,
				)
			if "invert_x" in settings and int(settings["invert_x"]):
				sens = -1.0 * sens[0], sens[1], sens[2]
			if "invert_y" in settings and int(settings["invert_y"]):
				sens = sens[0], -1.0 * sens[1], sens[2]
			if "invert_z" in settings and int(settings["invert_z"]):
				sens = sens[0], sens[1], -1.0 * sens[2]

			if sens != (1.0, 1.0, 1.0):
				action = SensitivityModifier(sens[0], sens[1], sens[2], action)

		return action

	@staticmethod
	def convert_key_name(name):
		"""Convert key names used in vdf profiles to Keys.KEY_* constants."""
		if name in VDFProfile.SPECIAL_KEYS:
			return VDFProfile.SPECIAL_KEYS[name]
		if name.endswith("_ARROW"):
			key = "KEY_%s" % (name[:-6],)
		elif "KEYPAD_" in name:
			key = "KEY_%s" % (name.replace("KEYPAD_", "KP"),)
		elif "LEFT_" in name:
			key = "KEY_%s" % (name.replace("LEFT_", "LEFT"),)
		elif "RIGHT_" in name:
			key = "KEY_%s" % (name.replace("RIGHT_", "RIGHT"),)
		else:
			key = "KEY_%s" % (name,)
		if hasattr(Keys, key):
			return getattr(Keys, key)
		if hasattr(Keys, key.upper()):
			return getattr(Keys, key.upper())
		raise ParseError("Unknown key: '%s'" % (name,))

	@staticmethod
	def convert_button_name(name):
		"""Convert button names used in vdf profiles to Keys.BTN_* constants."""
		if name.lower() in VDFProfile.SPECIAL_BUTTONS:
			return VDFProfile.SPECIAL_BUTTONS[name.lower()]
		key = "BTN_%s" % (name.upper(),)
		if hasattr(Keys, key):
			return getattr(Keys, key)
		raise ParseError("Unknown button: '%s'" % (name,))

	def parse_button(self, bdef, button=None):
		"""Parses button definition from vdf file.
		Parameter can be either string, as used in v2, or dict used in v3.
		"""
		if type(bdef) == str:
			# V2
			return self.parse_action(bdef, button)
		if type(bdef) == list:
			# V2
			return MultiAction.make(*[self.parse_action(x, button) for x in bdef])
		if "activators" in bdef:
			# V3
			act_actions = []
			for k in ("Full_Press", "Double_Press", "Long_Press"):
				a = NoAction()
				if k in bdef["activators"]:
					# TODO: Handle multiple bindings
					bindings = ensure_list(bdef["activators"][k])[0]
					a = self.parse_action(bindings["bindings"]["binding"], button)
					a = VDFProfile.parse_modifiers(bindings, a, Profile.RIGHT)
					# holly...
				act_actions.append(a)
			normal, double, hold = act_actions
			if not double and not hold:
				return normal
			if hold and not double:
				return HoldModifier(hold, normal)
			action = DoubleclickModifier(double, normal)
			action.holdaction = hold
			return action
		log.warning("Failed to parse button definition: %s" % (bdef,))

	@staticmethod
	def get_inputs(group):
		"""Returns 'inputs' or 'bindings', whichever exists in passed group.
		If neither exists, return None.
		"""
		if "inputs" in group:
			return group["inputs"]
		if "bindings" in group:
			return group["bindings"]
		return {}

	@staticmethod
	def find_group(data, id):
		"""Returns group with specified ID or None"""
		for g in data.get_all_for("group"):
			if "id" in g and g["id"] == id:
				return g
		return None

	def parse_group(self, group, side):
		"""Parses output (group) from vdf profile.
		Returns Action.
		"""
		if "mode" not in group:
			raise ParseError("Group without mode")
		mode = group["mode"]
		inputs = VDFProfile.get_inputs(group)

		settings = group["settings"] if "settings" in group else {}
		for o in ("output_trigger", "output_joystick"):
			if o in settings:
				if int(settings[o]) <= 1:
					side = Profile.LEFT
				else:
					side = Profile.RIGHT

		if mode == "dpad":
			keys = []
			for k in ("dpad_north", "dpad_south", "dpad_east", "dpad_west"):
				if k in inputs:
					keys.append(self.parse_button(inputs[k]))
				else:
					keys.append(NoAction())
			action = DPadAction(*keys)
		elif mode == "four_buttons":
			keys = []
			for k in ("button_y", "button_a", "button_x", "button_b"):
				if k in inputs:
					keys.append(self.parse_button(inputs[k]))
				else:
					keys.append(NoAction())
			action = DPadAction(*keys)
		elif mode == "joystick_move":
			if side == Profile.LEFT:
				# Left
				action = XYAction(AxisAction(Axes.ABS_X), AxisAction(Axes.ABS_Y))
			else:
				# Right
				action = XYAction(AxisAction(Axes.ABS_RX), AxisAction(Axes.ABS_RY))
		elif mode == "joystick_camera":
			output_joystick = 0
			if "output_joystick" in settings:
				output_joystick = int(settings["output_joystick"])
			if output_joystick == 0:
				action = BallModifier(XYAction(AxisAction(Axes.ABS_X), AxisAction(Axes.ABS_Y)))
			elif output_joystick == 1:
				action = BallModifier(XYAction(AxisAction(Axes.ABS_RX), AxisAction(Axes.ABS_RY)))
			else:
				# TODO: Absolute mouse? Doesn't seems to do anything in Steam
				action = BallModifier(SensitivityModifier(0.1, 0.1, MouseAction()))
		elif mode == "mouse_joystick":
			action = BallModifier(XYAction(AxisAction(Axes.ABS_RX), AxisAction(Axes.ABS_RY)))
		elif mode == "scrollwheel":
			action = BallModifier(XYAction(MouseAction(Rels.REL_HWHEEL), MouseAction(Rels.REL_WHEEL)))
		elif mode == "touch_menu":
			# Touch menu is converted to GridMenu
			items = []
			next_item_id = 1
			for k in inputs:
				action = self.parse_button(inputs[k])
				items.append(
					MenuItem(
						"item_%s" % (next_item_id,),
						action.describe(Action.AC_BUTTON),
						action,
					),
				)
				next_item_id += 1
			# Menu is stored in profile, with generated ID
			menu_id = "menu_%s" % (self.next_menu_id,)
			self.next_menu_id += 1
			self.menus[menu_id] = MenuData(*items)

			action = GridMenuAction(
				menu_id,
				"LEFT" if side == Profile.LEFT else "RIGHT",
				SCButtons.LPAD if side == Profile.LEFT else SCButtons.RPAD,
			)
		elif mode == "radial_menu":
			items = []
			next_item_id = 1
			for k in inputs:
				action = self.parse_button(inputs[k])
				items.append(MenuItem("item_%s" % (next_item_id,), action.describe(Action.AC_BUTTON), action))
				next_item_id += 1
			# Menu is stored in profile, with generated ID
			menu_id = "menu_%s" % (self.next_menu_id,)
			self.next_menu_id += 1
			self.menus[menu_id] = MenuData(*items)

			action = RadialMenuAction(
				menu_id,
				"LEFT" if side == Profile.LEFT else "RIGHT",
				SCButtons.LPAD if side == Profile.LEFT else SCButtons.RPAD,
			)
		elif mode == "absolute_mouse":
			if "click" in inputs:
				if side == Profile.LEFT:
					self.add_by_binding(SCButtons.LPAD, self.parse_button(inputs["click"]))
				else:
					self.add_by_binding(SCButtons.RPAD, self.parse_button(inputs["click"]))
			if "gyro_axis" in settings:
				if int(settings["gyro_axis"]) == 1:
					action = MouseAction(ROLL)
				else:
					action = MouseAction(YAW)
			else:
				action = MouseAction()
		elif mode == "mouse_wheel":
			action = BallModifier(XYAction(MouseAction(Rels.REL_HWHEEL), MouseAction(Rels.REL_WHEEL)))
		elif mode == "trigger":
			actions = []
			if "click" in inputs:
				actions.append(TriggerAction(TRIGGER_CLICK, self.parse_button(inputs["click"])))

			if side == Profile.LEFT:
				actions.append(AxisAction(Axes.ABS_Z))
			else:
				actions.append(AxisAction(Axes.ABS_RZ))

			action = MultiAction.make(*actions)
		elif mode == "mouse_region":
			# Read value and assume dafaults
			scale = float(settings["scale"]) if "scale" in settings else 100.0
			x = float(settings["position_x"]) if "position_x" in settings else 50.0
			y = float(settings["position_y"]) if "position_y" in settings else 50.0
			w = float(settings["sensitivity_horiz_scale"]) if "sensitivity_horiz_scale" in settings else 100.0
			h = float(settings["sensitivity_vert_scale"]) if "sensitivity_vert_scale" in settings else 100.0
			# Apply scale
			w = w * scale / 100.0
			h = h * scale / 100.0
			# Convert to (0, 1) range
			x, y = x / 100.0, 1.0 - (y / 100.0)
			w, h = w / 100.0, h / 100.0
			# Convert to rectangle
			x1 = max(0.0, x - (w * VDFProfile.REGION_IMPORT_FACTOR))
			x2 = min(1.0, x + (w * VDFProfile.REGION_IMPORT_FACTOR))
			y1 = max(0.0, y - (h * VDFProfile.REGION_IMPORT_FACTOR))
			y2 = min(1.0, y + (h * VDFProfile.REGION_IMPORT_FACTOR))

			action = RelAreaAction(x1, y1, x2, y2)
		else:
			raise ParseError("Unknown mode: '%s'" % (group["mode"],))

		action = VDFProfile.parse_modifiers(group, action, side)
		return action

	def parse_switches(self, group):
		"""Use for special cases of input groups that contain buttons."""
		inputs = VDFProfile.get_inputs(group)
		for button in inputs:
			if button in ("trigger_left", "left_trigger"):
				self.add_by_binding(
					"left_trigger",
					AxisAction(Axes.ABS_Z),
				)
			elif button in ("trigger_right", "right_trigger"):
				self.add_by_binding(
					"right_trigger",
					AxisAction(Axes.ABS_RZ),
				)
			elif button in VDFProfile.BUTTON_TO_BUTTON:
				self.add_by_binding(VDFProfile.BUTTON_TO_BUTTON[button], self.parse_button(inputs[button], button))
			else:
				raise ParseError("Unknown button: '%s'" % (button,))

	def parse_input_binding(self, data, group_id, binding):
		group = VDFProfile.find_group(data, group_id)
		if group and "mode" in group:
			if binding.startswith("switch") or binding.startswith("button_diamond"):
				self.parse_switches(group)
			else:
				if binding.startswith("right_"):
					action = self.parse_group(group, Profile.RIGHT)
				else:
					action = self.parse_group(group, Profile.LEFT)
				if binding.endswith("modeshift"):
					modeshift_id = (group_id, binding.split(" ")[0])
					self.modeshifts[modeshift_id] = action
				else:
					self.set_by_binding(binding, action)

	def set_by_binding(self, binding, action):
		"""Set action specified by binding, one of group_source_bindings keys used in vdf profile. Also supports SCButtons constants for buttons.

		Throws ParseError if key is not supported.
		"""
		if binding in SCButtons.__members__.values():
			self.buttons[binding] = action
		elif binding.startswith("left_trackpad"):
			self.pads[Profile.LEFT] = action
		elif binding.startswith("right_trackpad"):
			self.pads[Profile.RIGHT] = action
		elif binding.startswith("left_trigger"):
			self.triggers[Profile.LEFT] = action
		elif binding.startswith("right_trigger"):
			self.triggers[Profile.RIGHT] = action
		elif binding.startswith("joystick"):
			self.stick = action
		elif binding.startswith("gyro"):
			self.gyro = action
		else:
			raise ParseError("Unknown group source binding: '%s'" % (binding,))

	def add_by_binding(self, binding, action):
		"""As set_by_binding, but if there is alrady action for specified binding set, creates MultiAction."""
		old = self.get_by_binding(binding)
		new = MultiAction.make(old, action)
		if isinstance(new, MultiAction):
			new = new.deduplicate()
		self.set_by_binding(binding, new)

	def get_by_binding(self, binding):
		"""Return action specified by binding, one of group_source_bindings keys used in vdf profile. Also supports SCButtons constants for buttons.

		Throws ParseError if key is not supported.
		"""
		if binding in SCButtons.__members__.values():
			return self.buttons[binding]
		if binding.startswith("left_trackpad"):
			return self.pads[Profile.LEFT]
		if binding.startswith("right_trackpad"):
			return self.pads[Profile.RIGHT]
		if binding.startswith("left_trigger"):
			return self.triggers[Profile.LEFT]
		if binding.startswith("right_trigger"):
			return self.triggers[Profile.RIGHT]
		if binding.startswith("joystick"):
			return self.stick
		if binding.startswith("gyro"):
			return self.gyro
		raise ParseError("Unknown group source binding: '%s'" % (binding,))

	@staticmethod
	def _load_preset(data, profile, preset):
		profile.modeshifts = {}
		profile.modeshift_buttons = {}
		if "group_source_bindings" not in preset:
			# Empty preset
			return

		gsb = preset["group_source_bindings"]
		for group_id in gsb:
			binding = gsb[group_id]
			if not binding.endswith("inactive"):
				profile.parse_input_binding(data, group_id, binding)

		if "switch_bindings" in preset:
			profile.parse_switches(preset["switch_bindings"])

		for b in profile.modeshift_buttons:
			if profile.modeshift_buttons[b] in profile.modeshifts:
				# Should be always
				modeshift = profile.modeshift_buttons[b]
			else:
				continue
			action = profile.modeshifts[modeshift]
			trash, binding = modeshift
			old = profile.get_by_binding(binding)
			profile.set_by_binding(
				binding,
				ModeModifier(
					b,
					action,
					old,
				),
			)

	@staticmethod
	def _get_preset_name(data, preset):
		"""Return name of specified preset."""
		name = preset["name"].lower()
		if "actions" in data and name in data["actions"]:
			name = data["actions"][name]["title"]
		return name

	def action_set_by_id(self, id):
		"""Return name of action set with specified id."""
		for s in self.action_sets:
			if self.action_sets[s].action_set_id == id:
				return s
		return None

	def load(self, filename):
		"""Load profile from vdf file. Returns self.

		May raise ValueError.
		"""
		with open(filename) as file:
			data = parse_vdf(file)
		self.load_data(data)

	def load_data(self, data) -> Self:
		if "controller_mappings" not in data:
			raise ValueError("Invalid profile file")
		data = data["controller_mappings"]
		if "title" in data:
			name = data["title"].strip()
			if name:
				self.name = name
		presets = ensure_list(data["preset"])
		for p in presets:
			id = int(p["id"])
			if id == 0:
				# Default profile
				VDFProfile._load_preset(data, self, p)
			else:
				aset = VDFProfile(VDFProfile._get_preset_name(data, p))
				aset.action_set_id = id
				aset.action_set_switches = self.action_set_switches
				self.action_sets[aset.name] = aset
				VDFProfile._load_preset(data, aset, p)

		for aset in self.action_sets.values():
			aset.buttons[SCButtons.C] = HoldModifier(
				MenuAction("Default.menu"),
				MenuAction("Default.menu"),
			)

		return self


if __name__ == "__main__":
	import sys

	from scc.tools import init_logging

	init_logging()
	f = VDFProfile().load(sys.argv[1])
	f.save("output.sccprofile")



import logging
import os
import time
import traceback

from scc.actions import ButtonAction, GyroAbsAction
from scc.aliases import ALL_AXES, ALL_BUTTONS
from scc.config import Config
from scc.constants import (
	CPAD,
	DPAD,
	FE_PAD,
	FE_STICK,
	FE_TRIGGER,
	LEFT,
	RIGHT,
	RSTICK,
	STICK,
	STICK_PAD_MAX,
	STICKTILT,
	ControllerFlags,
	HapticPos,
	SCButtons,
)
from scc.controller import HapticData
from scc.lib import xwrappers as X
from scc.uinput import Dummy, Keyboard, Mouse, UInput

log = logging.getLogger("Mapper")


class Mapper:
	DEBUG = False

	def __init__(
		self,
		profile,
		scheduler,
		keyboard=b"SCController Keyboard",
		mouse=b"SCController Mouse",
		gamepad=True,
		poller=None,
	):
		"""If any of keyboard, mouse or gamepad is set to None, that device will not be emulated.

		Emulated gamepad will have rumble enabled only if poller is set to instance and configuration allows it.
		"""
		self.profile = profile
		self.controller = None
		self.xdisplay = None
		self.scheduler = scheduler

		# Create virtual devices
		log.debug("Creating virtual devices")
		self.keyboard = self.create_keyboard(keyboard) if keyboard else Dummy()
		log.debug("Keyboard: %s" % (self.keyboard,))
		self.mouse = self.create_mouse(mouse) if mouse else Dummy()
		log.debug("Mouse:    %s" % (self.mouse,))
		self.gamepad = self.create_gamepad(gamepad, poller) if gamepad else Dummy()
		log.debug("Gamepad:  %s" % (self.gamepad,))

		# Set by SCCDaemon instance; Used to handle actions
		# from scc.special_actions
		self._sa_handler = None

		# Setup emulation
		self.keypress_list = []
		self.keyrelease_list = []
		self.mouse_movements = [0,0,0,0,0,0]  # mouse x, y, wheel vertical, horisontal, stick mouse x, stick mouse y
		self.feedbacks = [None, None]  # left, right
		self.pressed = {}  # for ButtonAction, holds number of times virtual button was pressed without releasing it first
		self.syn_list = set()
		self.buttons, self.old_buttons = 0, 0
		self.lpad_touched = False
		self.state, self.old_state = None, None
		self.force_event = set()
		self.time_elapsed = 0.0

	def create_gamepad(self, enabled, poller):
		"""Parses gamepad configuration and creates apropriate unput device"""
		if not enabled or "SCC_NOGAMEPAD" in os.environ:
			# Completly undocumented and for debuging purposes only.
			# If set, no gamepad is emulated
			self.gamepad = Dummy()
			return None
		cfg = Config()
		keys = ALL_BUTTONS[0 : cfg["output"]["buttons"]]
		vendor = int(cfg["output"]["vendor"], 16)
		product = int(cfg["output"]["product"], 16)
		version = int(cfg["output"]["version"], 16)
		name = cfg["output"]["name"]
		rumble = cfg["output"]["rumble"] and poller != None
		axes: list[tuple[int, int, int, int, int]] = []
		i = 0
		for min, max in cfg["output"]["axes"]:
			fuzz, flat = 0, 0
			if abs(max - min) > STICK_PAD_MAX:
				fuzz, flat = 16, 128
			try:
				axes.append((ALL_AXES[i], min, max, fuzz, flat))
			except IndexError:
				# Out of axes
				break
			i += 1

		ui = UInput(
			vendor=vendor, product=product, version=version, name=name, keys=keys, axes=axes, rels=[], rumble=rumble,
		)
		if poller and rumble:
			poller.register(ui.getDescriptor(), poller.POLLIN, self._rumble_ready)
		return ui

	def create_keyboard(self, name):
		return Keyboard(name=name)

	def create_mouse(self, name):
		return Mouse(name=name)

	def _rumble_ready(self, fd, event):
		# Taken from Steam Controller Singer project
		# https://gitlab.com/Pilatomic/SteamControllerSinger
		STEAM_CONTROLLER_MAGIC_PERIOD_RATIO = 495483.0
		ef = self.gamepad.ff_read()
		if ef:  # tale of...
			if self.controller and self.controller.get_type() != "sc":
				amplitude = min(abs(ef.level), 0x7FFF) if ef.repetitions > 0 else 0
				period_command = 1024
				duration_seconds = ef.duration / 1000.0 * ef.repetitions
				count = min(round(duration_seconds * 0x10000 / period_command), 0x7FFF) if amplitude else 0
				self.send_feedback(
					HapticData(HapticPos.BOTH, period=period_command, amplitude=amplitude, count=count),
				)
				self.generate_feedback()
				return

			period_command = 0
			amplitude = 0
			if ef.level != 0 and ef.repetitions > 0:
				tempRatio = ef.level / 32767.5
				period_command = (6000 - 25000) * tempRatio + 25000
				amplitude = (900 - 600) * tempRatio + 600

			raw_period = period_command / STEAM_CONTROLLER_MAGIC_PERIOD_RATIO
			# duration_seconds = 1
			duration_seconds = ef.duration / 1000.0 * ef.repetitions
			count = 0
			if raw_period != 0:
				count = min(int(duration_seconds * 1.5 / raw_period), 0x7FFF)

			# log.debug(f"{ef.level} {ef.duration} {ef.repetitions} {count}")
			self.send_feedback(
				HapticData(
					HapticPos.BOTH,
					period=period_command,
					amplitude=amplitude,
					count=count,
					# period = 20000,
					# amplitude = max(0, ef.level),
					# count = min(0x7FFF, ef.duration * ef.repetitions / 30)
				),
			)
			self.generate_feedback()

	def get_gamepad_name(self):
		"""Returns name of emulated gamepad (as displayed by jstest & co)
		or None if Dummy is assigned.
		"""
		if isinstance(self.gamepad, Dummy):
			return None
		return self.gamepad.name

	def sync(self):
		"""Syncs generated events"""
		if len(self.syn_list):
			for dev in self.syn_list:
				dev.synEvent()
			self.syn_list = set()

	def set_controller(self, c):
		"""Sets controller device, used by some (one so far) actions"""
		self.controller = c

	def get_controller(self):
		"""Returns assigned controller device or None if no controller is set"""
		return self.controller

	def set_special_actions_handler(self, sa):
		self._sa_handler = sa

	def get_special_actions_handler(self):
		return self._sa_handler

	def set_xdisplay(self, x):
		self.xdisplay = x

	def get_xdisplay(self):
		return self.xdisplay

	def get_current_window(self):
		"""Returns window id of current window or None if xdisplay is not set
		"""
		if self.xdisplay:
			return X.get_current_window(self.xdisplay)
		return None

	def schedule(self, delay, cb):
		"""Schedules callback to be ran no sooner than after delay.
		Delay is float number in seconds.
		Callback is called with mapper as only argument.
		"""
		return self.scheduler.schedule(delay, cb, self)

	def cancel_task(self, task):
		"""Removes scheduled task."""
		return self.scheduler.cancel_task(task)

	def mouse_move(self, dx, dy):
		"""Schedules mouse movement to be done at end of processing callback.
		Called from actions while callback is being processed.
		"""
		self.mouse_movements[0] += dx
		self.mouse_movements[1] += dy

	def mouse_wheel(self, wx, wy):
		"""Schedules mouse wheel movement to be done at end of processing callback.
		Called from actions while callback is being processed.
		"""
		self.mouse_movements[2] += wx
		self.mouse_movements[3] += wy

	def mouse_move_stick(self, dx, dy):
		"""Schedules mouse movement to be done at end of processing callback.
		Called from actions while callback is being processed.
		"""
		self.mouse_movements[4] += dx
		self.mouse_movements[5] += dy

	def send_feedback(self, hapticdata):
		"""Schedules haptic feedback to be sent at end of processing callback.
		Called from actions while callback is being processed.
		"""
		if hapticdata.get_position() == HapticPos.BOTH:
			# HapticPos.BOTH is special case as controller doesn't
			# really support doing that by itself.
			self.feedbacks[0] = hapticdata.with_position(HapticPos.LEFT)
			self.feedbacks[1] = hapticdata.with_position(HapticPos.RIGHT)
		else:
			self.feedbacks[hapticdata.get_position()] = hapticdata

	def controller_flags(self):
		"""Returns controller flags or, if there is no controller set to
		this mapper, sc_by_cable driver matching defaults.
		"""
		return 0 if self.controller is None else self.controller.flags

	def is_touched(self, what) -> bool:
		"""Returns True if specified pad is being touched.
		May randomly return False for aphephobic pads.

		'what' should be LEFT or RIGHT (from scc.constants)
		"""
		if what == LEFT:
			return self.buttons & SCButtons.LPADTOUCH
		if what == RIGHT:
			return self.buttons & SCButtons.RPADTOUCH
		if what == CPAD:
			return self.buttons & SCButtons.CPADTOUCH
		return False

	def was_touched(self, what):
		"""As is_touched, but returns True if pad *was* touched
		in previous known state.

		This is used as:
		is_touched() and not was_touched() -> pad was just pressed
		not is_touched() and was_touched() -> pad was just released
		"""
		if what == LEFT:
			return self.old_buttons & SCButtons.LPADTOUCH
		if what == RIGHT:
			return self.old_buttons & SCButtons.RPADTOUCH
		if what == CPAD:
			return self.old_buttons & SCButtons.CPADTOUCH
		return False

	def is_pressed(self, button):
		"""Returns True if button is pressed
		"""
		if button == LEFT:
			button = SCButtons.LPAD
		elif button == RIGHT:
			button = SCButtons.RPAD
		return self.buttons & button

	def was_pressed(self, button):
		"""Returns True if button was pressed in previous known state
		"""
		if button == LEFT:
			button = SCButtons.LPAD
		elif button == RIGHT:
			button = SCButtons.RPAD
		return self.old_buttons & button

	def get_pressed_button(self):
		"""Gets button that was pressed by very last handled event or None,
		if last event doesn't involved button pressing.
		"""
		for x in SCButtons:
			if x & self.buttons & ~self.old_buttons:
				return x
		return None

	def set_button(self, button, state):
		"""Sets button state on input.
		Set value will stay only for durration of one event loop iteration.

		Used _temporarely_ by RingAction to emulate finger lifting from pad.
		"""
		if button == LEFT:
			button = SCButtons.LPADTOUCH
		elif button == RIGHT:
			button = SCButtons.RPADTOUCH

		if state:
			self.buttons |= button
		else:
			self.buttons &= ~button

	def set_was_pressed(self, button, state):
		"""As set_button, but changes value remembered
		from loop iteration before current.

		Used _temporarely_ by RingAction to emulate finger lifting from pad.
		"""
		if button == LEFT:
			button = SCButtons.LPADTOUCH
		elif button == RIGHT:
			button = SCButtons.RPADTOUCH

		if state:
			self.old_buttons |= button
		else:
			self.old_buttons &= ~button

	def release_virtual_buttons(self):
		"""Called when daemon is killed or USB dongle is disconnected.
		Sends button release event for every virtual button that is still being
		pressed.
		"""
		to_release, self.pressed = self.pressed, {}
		for x in to_release:
			ButtonAction._button_release(self, x, True)

	def cancel_all(self):
		"""Called when profile is changed to let all actions to cancel
		long-running effects they may have created
		"""
		for a in self.profile.get_actions():
			a.cancel(self)

	def reset_gyros(self):
		for a in self.profile.get_all_actions():
			if isinstance(a, GyroAbsAction):
				a.reset()

	def input(self, controller, old_state, state):
		# print(type(controller), type(old_state), type(state))
		# Store states
		self.old_state = old_state
		self.old_buttons = self.buttons

		self.state = state
		self.buttons = state.buttons

		t = time.time()
		controller.time_elapsed = self.time_elapsed = t - controller.lastTime
		controller.lastTime = t

		if self.buttons & SCButtons.LPAD and not self.buttons & (SCButtons.LPADTOUCH | STICKTILT):
			self.buttons = (self.buttons & ~SCButtons.LPAD) | SCButtons.STICKPRESS

		fe = self.force_event
		self.force_event = set()

		# Check buttons
		xor = self.old_buttons ^ self.buttons
		btn_rem = xor & self.old_buttons
		btn_add = xor & self.buttons

		try:
			if btn_add or btn_rem:
				# At least one button was pressed
				for x in self.profile.buttons:
					if x & btn_add:
						self.profile.buttons[x].button_press(self)
					elif x & btn_rem:
						self.profile.buttons[x].button_release(self)

			# Check sticks
			if self.controller.flags & ControllerFlags.SEPARATE_STICK:
				if FE_STICK in fe or self.old_state.stick_x != state.stick_x or self.old_state.stick_y != state.stick_y:
					self.profile.stick.whole(self, state.stick_x, state.stick_y, STICK)
			elif not self.buttons & SCButtons.LPADTOUCH:
				if FE_STICK in fe or self.old_state.lpad_x != state.lpad_x or self.old_state.lpad_y != state.lpad_y:
					self.profile.stick.whole(self, state.lpad_x, state.lpad_y, STICK)
			if self.controller.flags & ControllerFlags.IS_DECK:
				if (
					FE_STICK in fe
					or self.old_state.rstick_x != state.rstick_x
					or self.old_state.rstick_y != state.rstick_y
				):
					self.profile.rstick.whole(self, state.rstick_x, state.rstick_y, RSTICK)

			# Check gyro
			if controller.get_gyro_enabled():
				self.profile.gyro.gyro(
					self, state.gpitch, state.gyaw, state.groll, state.q1, state.q2, state.q3, state.q4,
				)

			# Check triggers
			if FE_TRIGGER in fe or state.ltrig != self.old_state.ltrig:
				if LEFT in self.profile.triggers:
					self.profile.triggers[LEFT].trigger(self, state.ltrig, self.old_state.ltrig)
			if FE_TRIGGER in fe or state.rtrig != self.old_state.rtrig:
				if RIGHT in self.profile.triggers:
					self.profile.triggers[RIGHT].trigger(self, state.rtrig, self.old_state.rtrig)

			# Check pads
			# RPAD
			if controller.flags & ControllerFlags.IS_DECK or controller.flags & ControllerFlags.HAS_RSTICK:
				if FE_PAD in fe or self.old_state.rpad_x != state.rpad_x or self.old_state.rpad_y != state.rpad_y:
					self.profile.pads[RIGHT].whole(self, state.rpad_x, state.rpad_y, RIGHT)
			elif FE_PAD in fe or self.buttons & SCButtons.RPADTOUCH or SCButtons.RPADTOUCH & btn_rem:
				self.profile.pads[RIGHT].whole(self, state.rpad_x, state.rpad_y, RIGHT)
			# DPAD
			if controller.flags & ControllerFlags.IS_DECK:
				if FE_PAD in fe or self.old_state.dpad_x != state.dpad_x or self.old_state.dpad_y != state.dpad_y:
					self.profile.pads[DPAD].whole(self, state.dpad_x, state.dpad_y, DPAD)

			# LPAD
			if self.controller.flags & ControllerFlags.SEPARATE_STICK:
				if FE_PAD in fe or self.old_state.lpad_x != state.lpad_x or self.old_state.lpad_y != state.lpad_y:
					self.profile.pads[LEFT].whole(self, state.lpad_x, state.lpad_y, LEFT)
			elif self.buttons & SCButtons.LPADTOUCH:
				# Pad is being touched now
				if not self.lpad_touched:
					self.lpad_touched = True
				self.profile.pads[LEFT].whole(self, state.lpad_x, state.lpad_y, LEFT)
				if self.old_state.buttons & STICKTILT and not self.buttons & STICKTILT:
					# LPAD and stick share axes and so when they are used simultaneously (by someone with 3 hands or so :)
					# this is how mapper can tell that stick was recentered
					self.profile.stick.whole(self, 0, 0, STICK)
			elif not self.buttons & STICKTILT:
				# Pad is not being touched
				if self.lpad_touched:
					self.lpad_touched = False
					self.profile.pads[LEFT].whole(self, 0, 0, LEFT)

			# CPAD (touchpad on DS4 controller)
			if controller.flags & ControllerFlags.HAS_CPAD:
				if (
					(FE_PAD in fe)
					or (self.old_state.cpad_x != state.cpad_x)
					or (self.old_state.cpad_y != state.cpad_y)
					or ((self.old_buttons & SCButtons.CPADTOUCH) and not (self.buttons & SCButtons.CPADTOUCH))
				):
					if self.buttons & SCButtons.CPADTOUCH:
						self.profile.pads[CPAD].whole(self, state.cpad_x, state.cpad_y, CPAD)
					elif self.old_buttons & SCButtons.CPADTOUCH:
						self.profile.pads[CPAD].whole(self, 0, 0, CPAD)
		except Exception:
			# Log error but don't crash here, it breaks too many things at once
			if hasattr(self, "_testing"):
				raise
			log.error("Error while processing controller event")
			log.error(traceback.format_exc())

		# TODO: Is it important to run scheduled stuff before generate_events?
		self.scheduler.run()
		self.generate_events()
		self.generate_feedback()

	def generate_events(self):
		# Generate events - keys
		if len(self.keypress_list):
			self.keyboard.pressEvent(self.keypress_list)
			self.keypress_list = []
		if len(self.keyrelease_list):
			self.keyboard.releaseEvent(self.keyrelease_list)
			self.keyrelease_list = []
		# Generate events - mouse
		mx, my, wx, wy, sx, sy = self.mouse_movements
		if mx != 0 or my != 0:
			self.mouse.moveEvent(int(mx), int(my * -1), self.time_elapsed)
			self.syn_list.add(self.mouse)
		if wx != 0 or wy != 0:
			self.mouse.scrollEvent(wx, wy)
			self.syn_list.add(self.mouse)
		if sx != 0 or sy != 0:
			# log.debug("STARTING")
			# log.debug(f"{sx} {sy}")
			self.mouse.moveStickEvent(sx, sy * -1, self.time_elapsed)
			self.syn_list.add(self.mouse)

		self.mouse_movements = [0, 0, 0, 0, 0, 0]
		self.sync()

	def generate_feedback(self):
		if self.controller:
			left, right = self.feedbacks
			if (
				self.controller.get_type() in ("ds4", "ds4evdev")
				and left
				and right
				and left.data[1:] == right.data[1:]
			):
				self.feedbacks = [None, None]
				self.controller.feedback(left.with_position(HapticPos.BOTH))
				return
			for x in (0, 1):
				if self.feedbacks[x]:
					self.controller.feedback(self.feedbacks[x])
					self.feedbacks[x] = None
